#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Tools to facilitate quantizeml models conversion to Akida pipelines.
"""

from ..akida_versions import AkidaVersion

# Mapper to match layers block patterns with the right conversion method, populated thanks to
# `register_conversion_patterns` function.
_V2_PATTERN_CONVERTERS = {}
_V1_PATTERN_CONVERTERS = {}
_V2_INPUT_PATTERN_CONVERTERS = {}
_V1_INPUT_PATTERN_CONVERTERS = {}


def register_conversion_patterns(version, patterns, converter, input_pattern=False):
    """Helper to populate the patterns_to_function dictionaries according to the right version.

    Args:
        version (AkidaVersion): the target Akida version.
        patterns (list): list of compatible quantized layers type patterns.
        converter (:obj:`BlockConverter`): the associated converter object.
        input_pattern (bool, optional): boolean to specify if the pattern should match an Akida
            input layer (such as a Stem or InputConv2D). Defaults to False.
    """
    if input_pattern:
        if version == AkidaVersion.v2:
            dict_ver = _V2_INPUT_PATTERN_CONVERTERS
        else:
            dict_ver = _V1_INPUT_PATTERN_CONVERTERS
    else:
        dict_ver = _V2_PATTERN_CONVERTERS if version == AkidaVersion.v2 else _V1_PATTERN_CONVERTERS
    for pattern in patterns:
        dict_ver.update({pattern: converter})


class BlockConverter:
    """This represents the main class that allows the conversion of a block of quantized layers with
    quantizeml into their Akida layer equivalent.

    It extracts the layers block pattern, checks if it's valid, applies optional additional checks
    (that should be implemented and integrated in the sub class constructor) and will allow the
    conversion of the block through the method convert.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        self._block = block

    def convert(self, model_ak):
        raise NotImplementedError

    @property
    def pattern(self):
        return [layer.__class__ for layer in self._block]

    @property
    def block_names(self):
        return [layer.name for layer in self._block]



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""BatchNormalization transformations for Keras/CNN2SNN Sequential models.
"""

import numpy as np
import tensorflow as tf
from keras import Sequential, Input
from keras.layers import (Conv2D, SeparableConv2D, Dense, MaxPool2D,
                          GlobalAvgPool2D, BatchNormalization)

from ..quantization_ops import MaxPerAxisQuantizer, StdPerAxisQuantizer
from ..quantization_layers import (QuantizedConv2D, QuantizedSeparableConv2D,
                                   QuantizedDense)
from ..cnn2snn_objects import cnn2snn_objects
from .clone import clone_layer, clone_layer_and_add_to_model


def invert_batchnorm_pooling(model):
    """Inverts pooling and BatchNormalization layers in a Sequential model to
    have BN layer before pooling.

    Having pool->BN or BN->pool is equivalent only if BN layer has no negative
    gammas.

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras model.
    Returns:
        :obj:`tf.keras.Model`: a Sequential Keras model.
    """

    assert isinstance(model, Sequential)

    new_model = Sequential()
    new_model.add(Input(model.input_shape[1:]))
    i = 0
    while i < len(model.layers) - 1:
        layer = model.layers[i]
        next_layer = model.layers[i + 1]
        if (isinstance(layer, (MaxPool2D, GlobalAvgPool2D))
                and isinstance(next_layer, BatchNormalization)):
            gammas = next_layer.get_weights()[0]
            if isinstance(layer, MaxPool2D) and np.any(gammas <= 0):
                # It is impossible to invert MaxPool->BN with gammas <= 0
                raise RuntimeError(f"There are {np.sum(gammas <= 0)} negative "
                                   "gammas in the batch norm layer "
                                   f"{next_layer.name}. Negative gammas are "
                                   "not supported.")
            # GlobalAveragePooling2D brings a change on axis for the batch norm.
            if isinstance(layer, GlobalAvgPool2D):
                bn_config = next_layer.get_config()
                bn_config['axis'] = [-1]
                with tf.keras.utils.custom_object_scope(cnn2snn_objects):
                    bn_layer_clone = BatchNormalization.from_config(bn_config)
            else:
                bn_layer_clone = clone_layer(next_layer)
            new_model.add(bn_layer_clone)
            bn_layer_clone.set_weights(next_layer.get_weights())
            clone_layer_and_add_to_model(layer, new_model)
            i = i + 2
        else:
            clone_layer_and_add_to_model(layer, new_model)
            i = i + 1

    if i < len(model.layers):
        clone_layer_and_add_to_model(model.layers[-1], new_model)

    return new_model


def _compute_BN_folded_weights(neural_layer, bn_layer):
    """Computes the new weights of a neural layer after folding BN layer.

    Args:
        neural_layer (:obj:`tf.keras.Layer`): a neural layer where BN will be
            folded.
        bn_layer (:obj:`tf.keras.Layer`): the BatchNormalization layer to fold
            into the neural layer.
    Returns:
        list: a list of the new weights to set in the new folded neural layer.
        list: a list of positive scale factors introduced by the folding.

    """

    # Get kernel and bias weights of the neural layer
    if type(neural_layer) in (SeparableConv2D, QuantizedSeparableConv2D):
        kernel_position = 1
        bias_position = 2
    else:
        kernel_position = 0
        bias_position = 1
    weights = neural_layer.get_weights()
    kernel = weights[kernel_position]
    bias = weights[bias_position] if neural_layer.use_bias else 0

    # Get BN weights
    gamma, beta, mean, var = bn_layer.get_weights()
    scale_BN = gamma / np.sqrt(var + bn_layer.epsilon)

    # Compute new folded kernel and bias
    new_kernel = kernel * scale_BN
    new_bias = beta + (bias - mean) * scale_BN

    # Return all weights with modified ones
    new_weights = weights
    new_weights[kernel_position] = new_kernel
    if neural_layer.use_bias:
        new_weights[bias_position] = new_bias
    else:
        new_weights.insert(bias_position, new_bias)

    # Absolute value of scale_BN is returned because we no longer need its sign.
    # It is later used to rescale the scale factors which are always positive.
    return new_weights, np.abs(scale_BN)


def fold_batchnorm(model):
    """Folds BatchNormalization layers into the preceding neural layers of
    a Sequential model.

    Args:
        model (:obj:`tf.keras.Model`): a Sequential Keras model.
    Returns:
        :obj:`tf.keras.Model`: a Sequential Keras model.
    """

    assert isinstance(model, Sequential)

    quantized_layers = (QuantizedConv2D, QuantizedSeparableConv2D,
                        QuantizedDense)
    neural_layers = quantized_layers + (Conv2D, SeparableConv2D, Dense)

    new_model = Sequential()
    new_model.add(Input(model.input_shape[1:]))
    i = 0
    while i < len(model.layers) - 1:
        layer = model.layers[i]
        next_layer = model.layers[i + 1]
        if (isinstance(layer, neural_layers) and isinstance(next_layer, BatchNormalization)):

            # Check BN axis parameter
            if (len(next_layer.axis) != 1 or next_layer.axis[0] != len(next_layer.input_shape) - 1):
                raise RuntimeError(f"The BatchNormalization layer "
                                   f"{next_layer.name} must be applied on the "
                                   f"last axis. Receives {next_layer.axis}.")

            # If the layer has been quantized, check quantizer
            if isinstance(layer, quantized_layers):
                if not isinstance(layer.quantizer,
                                  (MaxPerAxisQuantizer, StdPerAxisQuantizer)):
                    shift_for_sepconv = isinstance(layer,
                                                   QuantizedSeparableConv2D)
                    w = layer.get_weights()[0 + shift_for_sepconv]
                    scale_factors = layer.quantizer.scale_factor(tf.constant(w))
                    if tf.rank(scale_factors) != 1:
                        raise RuntimeError(
                            f"The BatchNormalization layer {next_layer.name} "
                            "can only be folded into a quantized layer that "
                            "uses a quantizer per axis.")

            # Add new neural layer with bias
            config = layer.get_config()
            config['use_bias'] = True
            new_layer = layer.__class__.from_config(config)
            new_model.add(new_layer)
            new_weights, scale_BN = _compute_BN_folded_weights(
                layer, next_layer)
            if np.any(scale_BN == 0):
                # Zero gammas are not supported: once folded, new kernel is zero
                raise RuntimeError(f"There are {np.sum(scale_BN == 0)} null "
                                   "gammas in the batch norm layer "
                                   f"{next_layer.name}. Null gammas are not "
                                   "supported.")
            new_layer.set_weights(new_weights)
            i = i + 2
        else:
            clone_layer_and_add_to_model(layer, new_model)
            i = i + 1

    if i < len(model.layers):
        clone_layer_and_add_to_model(model.layers[-1], new_model)

    return new_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Helper functions to convert QuantizedConv2D to its equivalent Akida layers.
"""
import numpy as np
from akida import PoolType
from quantizeml.layers import (QuantizedGlobalAveragePooling2D, QuantizedMaxPool2D,
                               QuantizedConv2DTranspose, QuantizedReLU, WeightQuantizer,
                               AlignedWeightQuantizer)
from .activations import parse_relu_v1, parse_relu_v2, set_relu_variables, v1_relu_checks
from .pooling import parse_max_pool_v1, parse_max_pool_v2
from .padding import get_padding, get_padding_value
from .weights import broadcast_and_set_variable
from .blocks import get_block_out_quantizer
from .outputs import set_output_v2_variables, parse_output_bits, parse_post_op_buffer_bits


def get_layer_by_type(block, type_list):
    """Helper to get the first layer, if found, of a block that match one of the types

    Args:
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.
        type_list (tuple): list of quantizeml layers types.

    Returns:
        :obj:`tf.keras.Layer`: the first quantized layer that match of the types of the list
        if found. None otherwise.
    """
    targeted_layer = None

    for layer in block:
        if isinstance(layer, type_list):
            targeted_layer = layer
            break

    return targeted_layer


def set_conv_variables_v1(layer_ak, layer_k, input_layer=False):
    """Computes and sets the variables for an Akida v1 Convolutional layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        layer_k (:obj:`tf.keras.Layer`): the source quantized layer.
        input_layer (bool, optional): Boolean to check if the targeted layer is a
            V1 conv input layer (i.e InputConvolutional). Defaults to False.

    """
    assert isinstance(layer_k.weight_quantizer, WeightQuantizer)
    if layer_k.use_bias:
        assert isinstance(layer_k.bias_quantizer, AlignedWeightQuantizer)

    # Get the weights
    weights = layer_k.weight_quantizer.qweights.value.fp.values.numpy()
    # Flip W and H dimensions for conv. kernels (not input conv.)
    if not input_layer:
        weights = np.flip(weights, axis=[0, 1])
    layer_ak.variables["weights"] = weights.astype(np.int8)

    # Set the bias (if there is one)
    if layer_k.use_bias:
        bias_quantizer = layer_k.bias_quantizer
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        # Store bias into the threshold variable
        layer_ak.variables["threshold"] = -bias


def set_conv_variables_v2(layer_ak, layer_k, input_layer=False):
    """Computes and sets the variables for an Akida v2 Conv2D layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        layer_k (:obj:`tf.keras.Layer`): the source quantized layer.
        input_layer (bool, optional): Boolean to check if the targeted layer is a
            V2 conv input layer (i.e InputConv2D). Defaults to False.
    """
    assert isinstance(layer_k.weight_quantizer, WeightQuantizer)
    if layer_k.use_bias:
        assert isinstance(layer_k.bias_quantizer, AlignedWeightQuantizer)

    # Get the weights
    weights = layer_k.weight_quantizer.qweights.value.fp.values.numpy()
    # Flip W and H dimensions for conv. kernels (not input conv.)
    if not input_layer:
        weights = np.flip(weights, axis=[0, 1])
    layer_ak.variables["weights"] = weights.astype(np.int8)

    # Set the bias (if there is one)
    if layer_k.use_bias:
        bias_quantizer = layer_k.bias_quantizer
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)
        # Unshift the bias and store it
        layer_ak.variables["bias"] = (bias >> bias_shift).astype(np.int8)
        # Also store the bias shift
        broadcast_and_set_variable(layer_ak.variables, "bias_shift", bias_shift)
        input_shift = getattr(layer_k, 'input_shift', None)
        if input_shift is not None:
            broadcast_and_set_variable(layer_ak.variables, "input_shift",
                                       input_shift.value.numpy().astype(np.uint8))

    # Set padding value (in case of InputConv2D)
    if input_layer:
        padding_value = get_padding_value(layer_k).astype("uint8")
        broadcast_and_set_variable(layer_ak.variables, "padding_value", padding_value)


def _parse_additional_layers_v1(block, block_params):
    """Parse conv block additional layers into the parameters of one for Akida v1.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the layers block.
        block_params (dict): the current block parameters.

    """
    # Identify the next layers

    relu_layer = get_layer_by_type(block, QuantizedReLU)
    pool_layer = get_layer_by_type(block, (QuantizedGlobalAveragePooling2D, QuantizedMaxPool2D))

    if isinstance(pool_layer, QuantizedMaxPool2D):
        pool_params = parse_max_pool_v1(pool_layer)
        block_params.update(pool_params)
    elif isinstance(pool_layer, QuantizedGlobalAveragePooling2D):
        pool_params = dict(pool_type=PoolType.Average)
        block_params.update(pool_params)
    if relu_layer:
        v1_relu_checks(relu_layer)
        act_params = parse_relu_v1(relu_layer)
        block_params.update(act_params)


def _parse_additional_layers_v2(block, block_params):
    """Parse conv block additional layers into the parameters of one for Akida v2.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the layers block.
        block_params (dict): the current block parameters.

    """
    # Identify the next layers

    relu_layer = get_layer_by_type(block, QuantizedReLU)
    pool_layer = get_layer_by_type(block, (QuantizedGlobalAveragePooling2D, QuantizedMaxPool2D))

    if isinstance(pool_layer, QuantizedMaxPool2D):
        pool_params = parse_max_pool_v2(pool_layer)
        block_params.update(pool_params)
    elif isinstance(pool_layer, QuantizedGlobalAveragePooling2D):
        block_params.update({"pool_type": PoolType.Average})
    if relu_layer:
        act_params = parse_relu_v2(relu_layer)
        block_params.update(act_params)
    # parse the block output bits
    parse_output_bits(block, block_params)
    # parse the block post op buffer bits
    parse_post_op_buffer_bits(block, block_params)


def parse_conv_block_v1(block, input_layer=False):
    """Parses a conv block parameters for an Akida v1 layer.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the conv block layers.
        input_layer (bool, optional): Boolean to check if the targeted layer is a
            V1 conv input layer (i.e InputConvolutional). Defaults to False.

    Returns:
        dict: the corresponding akida parameters.
    """
    conv = block[0]
    conv_params = dict(
        padding=get_padding(conv),
        kernel_size=(conv.kernel_size[0], conv.kernel_size[1]),
        filters=int(conv.kernel.shape[3]),
        weights_bits=conv.weight_quantizer.bitwidth,
        kernel_stride=(conv.strides[0], conv.strides[1]),
        activation=False,
        name=conv.name
    )
    # If the targeted layer is a InputConvolutional add two params.
    if input_layer:
        input_layer_params = dict(input_shape=tuple(int(x) for x in conv.input_shape[1:4]),
                                  padding_value=get_padding_value(conv))
        conv_params.update(input_layer_params)

    _parse_additional_layers_v1(block, conv_params)
    return conv_params


def parse_conv_block_v2(block, input_layer=False):
    """Parses a conv block parameters for akida v2.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the conv block layers.
        input_layer (bool, optional): Boolean to check if the targeted layer is a
            V2 conv input layer (i.e InputConv2D). Defaults to False.

    Returns:
        dict: the corresponding akida parameters.
    """
    conv = block[0]
    # Padding value must be built in constructor
    padding = get_padding(conv)

    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    conv_params = dict(
        padding=padding,
        kernel_size=conv.kernel_size[0],
        filters=int(conv.kernel.shape[3]),
        kernel_stride=conv.strides[0],
        activation=False,
        buffer_bits=conv.buffer_bitwidth + 1,
        name=conv.name
    )
    # If the targeted layer is a InputConv2D add two params.
    if input_layer:
        conv_params["input_shape"] = tuple(conv.input_shape[1:4])

    _parse_additional_layers_v2(block, conv_params)
    return conv_params


def parse_sepconv_block(block):
    """Parses a quantizeml sepconv block parameters for Akida v1.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the sepconv block layers.

    Returns:
        dict: the corresponding akida parameters.
    """
    sepconv = block[0]
    # Padding value must be built in constructor
    padding = get_padding(sepconv)

    sepconv_params = dict(kernel_size=(sepconv.kernel_size[0], sepconv.kernel_size[1]),
                          filters=int(sepconv.pointwise_kernel.shape[3]),
                          padding=padding,
                          kernel_stride=(sepconv.strides[0], sepconv.strides[1]),
                          weights_bits=sepconv.dw_weight_quantizer.bitwidth,
                          activation=False,
                          name=sepconv.name)

    _parse_additional_layers_v1(block, sepconv_params)
    return sepconv_params


def parse_depthwise_conv_block(block):
    """Parses a quantizeml depthwise conv block parameters for Akida v2.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the depthwise conv2d block layers.

    Returns:
        dict: the corresponding akida parameters.
    """
    depth_conv = block[0]

    # Padding value must be built in constructor
    padding = get_padding(depth_conv)

    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    block_params = dict(
        kernel_size=depth_conv.kernel_size[0],
        kernel_stride=depth_conv.strides[0],
        padding=padding,
        buffer_bits=depth_conv.buffer_bitwidth + 1,
        activation=False,
        name=depth_conv.name
    )
    _parse_additional_layers_v2(block, block_params)

    return block_params


def parse_conv2d_transpose_block(block, depthwise=False):
    """Parses a quantizeml (depthwise)conv transpose block parameters for Akida v2.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the (depthwise)conv2d_transpose block layers.
        depthwise (bool, optional): boolean to declare the main layer as a depthwise layer.

    Returns:
        dict: the corresponding akida parameters.
    """

    conv_transpose = block[0]

    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    buffer_bits = conv_transpose.buffer_bitwidth + 1

    block_params = dict(
        kernel_size=conv_transpose.kernel_size[0],
        buffer_bits=buffer_bits,
        activation=False,
        name=conv_transpose.name
    )
    # Add filters parameters if not a Depthwise convolution
    if not depthwise:
        assert isinstance(conv_transpose, QuantizedConv2DTranspose)
        block_params["filters"] = conv_transpose.filters

    relu_layer = get_layer_by_type(block, QuantizedReLU)
    if relu_layer:
        act_params = parse_relu_v2(relu_layer)
        block_params.update(act_params)

    # parse the block output bits
    parse_output_bits(block, block_params)
    # parse the block post op buffer bits
    parse_post_op_buffer_bits(block, block_params)

    return block_params


def set_conv2d_transpose_block_variables(layer_ak, block):
    """Computes and sets the variables for an Akida v2 Conv2DTranspose or DepthwiseConv2DTranspose
    layers.

    This function converts the variables of a Keras layers block and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        block (list(:obj:`tf.keras.Layer`)): the (depthwise)conv2d_transpose block layers.
    """
    conv_transpose = block[0]

    assert isinstance(conv_transpose.weight_quantizer, WeightQuantizer)
    if conv_transpose.use_bias:
        assert isinstance(conv_transpose.bias_quantizer, AlignedWeightQuantizer)

    # Get the weights (Note that in qweights the filter and channel dimensions are already
    # transposed)
    weights = conv_transpose.weight_quantizer.qweights.value.fp.values.numpy()

    layer_ak.variables["weights"] = weights.astype(np.int8)

    # Set the bias (if there is one)
    if conv_transpose.use_bias:
        bias_quantizer = conv_transpose.bias_quantizer
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)

        # Unshift the bias and store it
        layer_ak.variables["bias"] = (bias >> bias_shift).astype(np.int8)
        # Also store the bias shift
        broadcast_and_set_variable(layer_ak.variables, "bias_shift", bias_shift)

    # Set input shift if available
    if getattr(conv_transpose, 'input_shift', None):
        broadcast_and_set_variable(layer_ak.variables, "input_shift",
                                   conv_transpose.input_shift.value.numpy().astype(np.uint8))

    # Check if we have ReLU
    relu_layer = get_layer_by_type(block, QuantizedReLU)
    # Set optional activation variables
    if relu_layer:
        set_relu_variables(layer_ak, relu_layer)

    # Get the layer block output_quantizer
    out_quantizer = get_block_out_quantizer(block)
    # Set optional output_quantizer variables
    if out_quantizer:
        set_output_v2_variables(layer_ak, out_quantizer)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert Dequantizer to Akida.
"""
import numpy as np

import akida
from quantizeml.layers import Dequantizer

from .weights import broadcast_and_set_variable
from .block_converter import BlockConverter, register_conversion_patterns
from ..akida_versions import AkidaVersion


__all__ = ["DequantizerBlockConverterV1", "DequantizerBlockConverterV2"]

_PATTERNS = [(Dequantizer,)]


def _get_dequantizer_scales(dequantizer):
    """Helper to extract the common Dequantizer layer scale variable.

    Args:
        dequantizer (:obj:`tf.keras.Layer`): the keras Dequantizer layer.

    Returns:
        ndarray: the layer scales in a numpy float32 array format.

    """
    assert isinstance(dequantizer, Dequantizer)

    # Extract the Dequantizer variables
    frac_bits = dequantizer.frac_bits
    scales = dequantizer.scales
    if isinstance(frac_bits, (list, tuple)):
        frac_bits = frac_bits[0]
        scales = scales[0] if scales else None

    # We project the frac_bits into the scales as:
    #   new_scales = scales / 2 ** frac_bits
    scales = scales.value if scales else 1.0
    scales /= 2**frac_bits.value
    scales = scales.numpy().astype(np.float32)

    return scales


def _convert_dequantizer_v1(model_ak, block):
    """Converts Dequantizer layer and set its variables into the Akida v1 model last layer.

    Args:
        model_ak (:obj:`akida.Model`): the model where the layer will be added.
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.

    Returns:
        bool: Returns True for a successful conversion.
    """

    # Take the last layer to set variables
    layer_ak = model_ak.layers[-1]

    if layer_ak.parameters.activation == 1:
        return True
    # Computes and sets the dequantizer recorded variables into the akida layer
    scales = _get_dequantizer_scales(block[0])
    layer_variables = layer_ak.variables
    # Project scales in the final layer activation step
    broadcast_and_set_variable(layer_variables, "act_step", layer_variables["act_step"] / scales)
    return True


def _convert_dequantizer_v2(model_ak, block):
    """Converts Dequantizer layer and set its variables into the Akida v2 model.

    Args:
        model_ak (:obj:`akida.Model`): the model where the layer will be added.
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.

    Returns:
        bool: Returns True for a successful conversion.
    """
    dequantizer = block[0]
    # Create and add one akida.Dequantizer to the model
    layer_ak = akida.Dequantizer(name=dequantizer.name)
    model_ak.add(layer_ak)
    # Computes and sets the dequantizer recorded variables into the akida layer
    scales = _get_dequantizer_scales(dequantizer)
    broadcast_and_set_variable(layer_ak.variables, "scales", scales)
    return True


def _block_additional_checks(block):
    """Additional checks done in the dequantizer layer block.

    Args:
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.
    """
    dequantizer = block[0]
    assert isinstance(dequantizer, Dequantizer)
    # Extract the Dequantizer variables
    frac_bits = dequantizer.frac_bits
    if isinstance(frac_bits, (list, tuple)):
        if len(frac_bits) > 1:
            raise RuntimeError(f"Multi-inbounds in {dequantizer.name} is not supported.")


class DequantizerBlockConverterV1(BlockConverter):
    """Main class that should be used to check if the dequantizer layer block is compatible to an
    Akida V1 conversion and provides a method to convert and integrate its variables in the last
    Akida model layer.

    Args:
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        _block_additional_checks(self._block)

    def convert(self, model_ak):
        return _convert_dequantizer_v1(model_ak, self._block)


class DequantizerBlockConverterV2(BlockConverter):
    """Main class that should be used to check if the dequantizer layer block is compatible to an
    Akida V2 conversion and provides a method to convert it in a corresponding Akida v2 Dequantizer
    layer.

    Args:
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        _block_additional_checks(self._block)

    def convert(self, model_ak):
        return _convert_dequantizer_v2(model_ak, self._block)


# Register the valid dequantizer block pattern for Akida v1 and v2
register_conversion_patterns(AkidaVersion.v1, _PATTERNS, DequantizerBlockConverterV1)
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, DequantizerBlockConverterV2)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedDense to Akida.
"""
from akida import Dense1D, Dense2D, FullyConnected
from quantizeml.layers import (QuantizedDense, QuantizedReLU, QuantizedReshape, QuantizedFlatten,
                               WeightQuantizer, AlignedWeightQuantizer)
import numpy as np

from .layer_utils import get_inbound_layers
from .weights import broadcast_and_set_variable
from ..akida_versions import AkidaVersion
from .activations import parse_relu_v1, parse_relu_v2, set_relu_variables, v1_relu_checks
from .outputs import (set_output_v1_variables, set_output_v2_variables, parse_output_bits,
                      parse_post_op_buffer_bits)
from .blocks import get_block_out_quantizer
from .block_converter import BlockConverter, register_conversion_patterns

__all__ = ["DenseBlockConverterV1", "DenseBlockConverterV2"]


_PATTERNS = [(QuantizedDense,), (QuantizedDense, QuantizedReLU), (QuantizedFlatten, QuantizedDense),
             (QuantizedFlatten, QuantizedDense, QuantizedReLU), (QuantizedReshape, QuantizedDense),
             (QuantizedReshape, QuantizedDense, QuantizedReLU),
             (QuantizedReshape, QuantizedFlatten, QuantizedDense),
             (QuantizedReshape, QuantizedFlatten, QuantizedDense, QuantizedReLU)]


def _set_dense_v1_variables(ak_layer, dense, block_input_shape):
    """Computes and sets the variables for an Akida FullyConnected layer.

    This function converts the variables of a dense block of Keras layers and sets them into
    the corresponding variables of the equivalent Akida FullyConnected layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        dense (:obj:`tf.keras.Layer`): the keras dense layer.
        block_input_shape (list): the dense block input shape.
    """
    assert isinstance(dense.weight_quantizer, WeightQuantizer)
    variables_ak = ak_layer.variables

    # get the QuantizedDense weights
    weights_ak = dense.weight_quantizer.qweights.value.fp.values.numpy()
    # get the QuantizedDense bias and shift
    if dense.use_bias:
        bias_quantizer = dense.bias_quantizer
        assert isinstance(bias_quantizer, AlignedWeightQuantizer)
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        # Store bias into the threshold variable
        variables_ak["threshold"] = -bias

    # In AkidaVersion.v1, it is only possible to convert flat inputs
    input_shape = dense.input_shape
    units = ak_layer.parameters.units
    if block_input_shape != input_shape:
        _, X, Y, C = block_input_shape
        # When a fully connected layer follows a convolutional layer, we need to modify
        # the way the weights are laid out because the Keras and Akida flatten operations
        # that happen on spatial dimensions are inverted (row-major versus col-major).
        # We therefore need to:
        # - reshape the Keras (FxN) weights to match the block input shape (XxYxCxN),
        # - transpose to the equivalent akida shape (CxYxXxN).
        weights_ak = weights_ak.reshape(X, Y, C, units).transpose(2, 1, 0, 3)
    weights_ak = weights_ak.reshape(1, 1, input_shape[-1], units)
    variables_ak["weights"] = weights_ak.astype(np.int8)


def _set_dense_v2_variables(ak_layer, dense):
    """Computes and sets the variables for an Akida Dense1D/Dense2D layer.

    This function converts the variables of a dense block of Keras layers and sets them into
    the corresponding variables of the equivalent Akida Dense1D/Dense2D layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        dense (:obj:`tf.keras.Layer`): the keras dense layer.
    """
    assert isinstance(dense.weight_quantizer, WeightQuantizer)
    variables_ak = ak_layer.variables

    # get the QuantizedDense weights
    weights_ak = dense.weight_quantizer.qweights.value.fp.values.numpy()
    # get the QuantizedDense bias and shift
    if dense.use_bias:
        bias_quantizer = dense.bias_quantizer
        assert isinstance(bias_quantizer, AlignedWeightQuantizer)
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)
        variables_ak["bias"] = (bias >> bias_shift).astype(np.int8)
        broadcast_and_set_variable(variables_ak, "bias_shift", bias_shift)

    input_shift = getattr(dense, 'input_shift', None)
    if input_shift is not None:
        broadcast_and_set_variable(variables_ak, "input_shift",
                                   input_shift.value.numpy().astype(np.uint8))

    variables_ak["weights"] = weights_ak.astype(np.int8)


def _parse_dense_v1(layer):
    """Parses a quantizeml.QuantizedDense parameters that target the akida v1 FullyConnected layer.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml.QuantizedDense layer to parse.

    Returns:
        dict: the corresponding akida parameters.

    """
    # Set by default the activation to false. It will be updated when parsing the optional ReLU
    # layer.
    dense_params = dict(units=layer.units, activation=False, name=layer.name)

    weight_bits = layer.weight_quantizer.bitwidth
    if weight_bits not in [1, 2, 4]:
        raise ValueError(f"FullyConnected layer only supports 1, 2 or 4 weight bits value. \
                              Received weight_bits={weight_bits}.")
    dense_params["weights_bits"] = weight_bits

    return dense_params


def _parse_dense_v2(layer):
    """Parses a quantizeml.QuantizedDense parameters that target the akida v2 Dense1D or Dense2D
    layer.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml.QuantizedDense layer to parse.

    Returns:
        dict: the corresponding akida parameters.

    """
    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    buffer_bits = layer.buffer_bitwidth + 1

    # Set by default the activation to false. It will be updated when parsing the optional ReLU
    # layer.
    dense_params = dict(units=layer.units, activation=False, name=layer.name)

    dense_params["buffer_bits"] = buffer_bits
    return dense_params


def _get_dense_layer(block):
    """Helper to extract QuantizedDense layer of the block.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the layers block to convert.

    Return:
        :obj:`tf.keras.Layer`: the QuantizedDense layer of the block
    """
    for layer in block:
        if isinstance(layer, QuantizedDense):
            return layer
    raise RuntimeError("No QuantizedDense found in that block."
                       f" Received: {[l.__class__.__name__ for l in block]}")


def convert_block_to_akida_dense_layer(model_ak, block, akida_type):
    """Converts a dense block into an akida dense layer target.

    The expected sequence is:
        - QuantizedFlatten/QuantizedReshape (optional),
        - QuantizedDense,
        - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the layer will be added.
        block (list(:obj:`tf.keras.Layer`)): the layers block to convert.
        akida_type (:obj:`akida.LayerType`): the targeted akida layer type.

    Return:
        bool: return True if success.
    """

    # Extract dense layer of the block
    dense = _get_dense_layer(block)

    # Parse main dense params
    if akida_type == FullyConnected:
        dense_params = _parse_dense_v1(dense)
    else:
        dense_params = _parse_dense_v2(dense)
        # parse the block output bits
        parse_output_bits(block, dense_params)
        # parse the block post op buffer bits
        parse_post_op_buffer_bits(block, dense_params)

    # Extract ReLU if present
    relu_layer = None
    if isinstance(block[-1], QuantizedReLU):
        relu_layer = block[-1]
        # parse ReLU layer parameters
        if akida_type == FullyConnected:
            act_params = parse_relu_v1(relu_layer)
        else:
            act_params = parse_relu_v2(relu_layer)
        dense_params.update(act_params)

    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, dense)

    # Create the Akida layer
    dense_ak = akida_type(**dense_params)
    model_ak.add(dense_ak, inbound_layers_ak)

    # Set the main Akida dense layer variables depending on the layer type
    if akida_type == FullyConnected:
        _set_dense_v1_variables(dense_ak, dense, block[0].input_shape)
    else:
        _set_dense_v2_variables(dense_ak, dense)
        # Set the optional activation variables for v2 layers:
        if relu_layer:
            set_relu_variables(dense_ak, relu_layer)

    # Get out_quantizer of the block.
    out_quantizer = get_block_out_quantizer(block)

    # Set the optional output_quantizer variables
    if out_quantizer:
        set_output_v1_variables(dense_ak, out_quantizer) if akida_type == FullyConnected else\
            set_output_v2_variables(dense_ak, out_quantizer)

    return True


class DenseBlockConverterV1(BlockConverter):
    """Main class that should be used to check if the dense block is compatible to an Akida v1
    conversion and provides a method to convert it in an equivalent Akida FullyConnected layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._dense_layer_checks()
        # Check ReLU compatibility for v1 only
        if isinstance(block[-1], QuantizedReLU):
            v1_relu_checks(block[-1])

    def convert(self, model_ak):
        return convert_block_to_akida_dense_layer(model_ak, self._block, FullyConnected)

    def _dense_layer_checks(self):
        dense = _get_dense_layer(self._block)
        # Get dense layer input shape without the batch size.
        dense_shape = dense.input_shape[1:]
        # For a shape (X, Y, Z), if X*Y*Z==Z <=> X=Y=1 <=> the shape is flattened.
        if not (np.prod(dense_shape) == dense_shape[-1]):
            raise RuntimeError("In v1 the dense layer only handles flattened shapes"
                               f" (i.e (1, 1, X) or (X,)). Received: ({dense.input_shape[1:]}) "
                               " shape.")


class DenseBlockConverterV2(BlockConverter):
    """Main class that should be used to check if the dense block is compatible to an Akida v2
    conversion and provides a method to convert it in an equivalent Akida Dens1D or Dense2D layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        dense = _get_dense_layer(self._block)
        # Get dense layer input shape without the Batch size.
        dense_shape = dense.input_shape[1:]
        # For a shape (X, Y, Z), if X*Y*Z==Z <=> X=Y=1 <=> the shape is flattened.
        self._flag_collapse_dims = (np.prod(dense_shape) == dense_shape[-1])

    def convert(self, model_ak):
        return convert_block_to_akida_dense_layer(model_ak, self._block, Dense1D) if \
            self._flag_collapse_dims else convert_block_to_akida_dense_layer(model_ak, self._block,
                                                                             Dense2D)


# Register the valid dense block patterns for Akida v1 and v2
register_conversion_patterns(AkidaVersion.v1, _PATTERNS, DenseBlockConverterV1)
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, DenseBlockConverterV2)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["Depthwise2DOnnxConverter"]

import akida

from .base_converter import OnnxConverter
from .check_compatibility import check_attributes
from .register import register_onnx_converter_target
from .conv_commons import set_convolutional_variables, check_convolution_compatibility
from .padding import get_akida_padding


@register_onnx_converter_target("QuantizedDepthwise2D")
class Depthwise2DOnnxConverter(OnnxConverter):
    """Convert QuantizedDepthwise2D type node into an akida.DepthwiseConv2D.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """

    def load_attributes(self, node):
        # Load default attributes
        super().load_attributes(node)

        # Some attributes should infer from node.op_type
        n_op = node.op_type
        self.activation = "ReLU" in n_op
        self.pool_type = akida.PoolType.NoPooling
        self.use_bias = "Biased" in n_op

        # Padding type is inferred from pads attribute
        self.pads = self.weights["pads"].tolist()
        self.padding = akida.Padding.Same if any(self.pads) else akida.Padding.Valid

    def _additional_checks(self):
        # Convolutional checks
        check_convolution_compatibility(self)

        # Group constrains
        check_attributes(self, ["groups"])
        filters = self.weights["W"].shape[0]
        channels = self.weights["W"].shape[1]
        if self.groups != filters or channels != 1:
            raise ValueError(f"The number of groups in the node {self.name} ({self.groups}) "
                             f"does not match the number of filters ({filters}).")

    def _parse_akida_layer(self):
        # Parse information
        layer_params = {
            "name": self.name,
            "activation": self.activation,
            "output_bits": 8,
            "kernel_size": self.weights["W"].shape[-1],
            "padding": get_akida_padding(self),
            "pool_type": self.pool_type,
            "kernel_stride": self.strides[0],
        }
        return akida.DepthwiseConv2D(**layer_params)

    def _set_akida_variables(self, ak_layer):
        set_convolutional_variables(self, ak_layer, is_depthwise=True)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["register_onnx_converter_target", "map_node_to_converter"]

from .base_converter import OnnxConverter

_ONNX_CONVERTERS = {}


def register_onnx_converter_target(root_op_type):
    """Register the current class as a target for converter modules.

    This decorator injects the decorated class into the _ONNX_CONVERTERS dictionary, so that
    it is registered to as converter target.

    Args:
        root_op_type (str): string that represents the base operation type
            which will map the decorator's target class.

    Returns:
        Callable: a decorator that registers the decorated class
    """
    def _register_target(target, arg):
        """Register arg with target name.

        Args:
            target (str): the target name to register
            arg (cls): the current class to register
        """
        if not issubclass(arg, OnnxConverter):
            raise ValueError("Can only register OnnxCoverter class objects.")

        if target in _ONNX_CONVERTERS:
            raise ValueError(f"{target} has already been registered to {_ONNX_CONVERTERS[target]}.")
        _ONNX_CONVERTERS.update({target: arg})

    def decorator(arg):
        assert isinstance(root_op_type, str)
        _register_target(root_op_type, arg)
        return arg
    return decorator


def map_node_to_converter(node, model):
    """Map node into its converter version.

    Args:
        node (NodeProto): the node to map.
        model (ModelProto): the model to initialize the converter.

    Raises:
        RuntimeError: when node is not register as a valid operation.

    Returns:
        OnnxConverter: the converter.
    """
    for root_op_type, converter in _ONNX_CONVERTERS.items():
        if node.op_type.startswith(root_op_type):
            return converter(node, model)
    raise RuntimeError(f"Unrecognized {node.op_type} operation. Not register yet!")



# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to generate block converters from a QuantizeML model.
"""
from keras import layers, Sequential
from quantizeml import layers as qlayers

from .blocks import split_model_into_blocks
from ..transforms.sequential import (_check_layers_data_format, _check_layers_dilation,
                                     _check_layer_inbounds)
from ..akida_versions import get_akida_version, AkidaVersion
from .block_converter import (_V1_PATTERN_CONVERTERS, _V2_PATTERN_CONVERTERS,
                              _V1_INPUT_PATTERN_CONVERTERS, _V2_INPUT_PATTERN_CONVERTERS)

neural_layers = (qlayers.QuantizedConv2D, qlayers.QuantizedSeparableConv2D,
                 qlayers.QuantizedDense, qlayers.QuantizedDepthwiseConv2D,
                 qlayers.QuantizedConv2DTranspose, qlayers.QuantizedDepthwiseConv2DTranspose,
                 qlayers.QuantizedBufferTempConv, qlayers.QuantizedDepthwiseBufferTempConv,
                 qlayers.QuantizedStatefulRecurrent)
skippable_layers = (layers.InputLayer, layers.Rescaling, layers.Activation, layers.Softmax,
                    layers.Dropout)
pooling_layers = (qlayers.QuantizedGlobalAveragePooling2D, qlayers.QuantizedMaxPool2D)
norm_layers = (qlayers.LayerMadNormalization, qlayers.QuantizedBatchNormalization)
activation_layers = (qlayers.QuantizedReLU,)
stem_layers = (qlayers.ClassToken, qlayers.AddPositionEmbs)
reshape_layers = (qlayers.QuantizedReshape, qlayers.QuantizedFlatten)


def _block_pattern(block):
    """Method that returns the pattern of a block of layers.

    Args:
        block (list): list of quantized quantizeml layers.

    Returns:
        tuple: list of layer types representing the block pattern.
    """
    return tuple([layer.__class__ for layer in block])


def _get_block_converter(block):
    """Helper to get the BlockConverter of a block of layers.

    Args:
        block (list): list of quantized quantizeml layers.

    Returns:
        (:obj:`BlockConverter`): the BlockConverter corresponding to the block of layers or None.
    """
    pattern = _block_pattern(block)

    if get_akida_version() == AkidaVersion.v1:
        block_converter = _V1_PATTERN_CONVERTERS.get(pattern, None)
    else:
        block_converter = _V2_PATTERN_CONVERTERS.get(pattern, None)

    return block_converter


def _get_input_block_converter(block):
    """Helper to get the BlockConverter of an input block of layers.

    Args:
        block (list): list of quantized quantizeml layers.

    Returns:
        (:obj:`BlockConverter`): the BlockConverter corresponding to the block of layers or None.
    """
    pattern = _block_pattern(block)

    if get_akida_version() == AkidaVersion.v1:
        return _V1_INPUT_PATTERN_CONVERTERS.get(pattern, None)
    return _V2_INPUT_PATTERN_CONVERTERS.get(pattern, None)


def _display_supported_patterns(layer_filter=None):
    """Helper to display the list of the supported patterns

    Args:
        layer_filter (list, optional): a layer list to limit the pattern list. Defaults to None.

    Returns:
        str: the list of supported patterns.
    """
    if get_akida_version() == AkidaVersion.v1:
        pattern_converters = _V1_PATTERN_CONVERTERS
    else:
        pattern_converters = _V2_PATTERN_CONVERTERS
    keys = list(pattern_converters.keys())
    display = ""
    for key in keys:
        layer_pattern = [layer.__name__ for layer in key]
        # Add the pattern if no filter or if there is a common layer
        if layer_filter is None or set(layer_pattern) & set(layer_filter):
            display += " - " + " > ".join(layer_pattern) + "\n"

    # if layer_filter filtered all patterns, display without filter
    if display == "" and len(keys) > 0:
        display = _display_supported_patterns()

    return display


def generate_block_converters(model):
    """Extract block converts from a QuantizeML model

    Exceptions are raised if incompatibilities encountered.

    Args:
        model (:obj:`tf.keras.Model`): the model to extract.

    Returns:
        list: a list of sequences of the non_skippable layers ('blocks').
    """
    # Check general rules about model in three steps:
    # 1. Check if model has only one input and one output,
    # 2. Check right data format and
    # 3. Over Akida 1.0, check if model is sequential.
    _check_model_input_output(model)
    _check_layers_data_format(model)
    _check_layers_dilation(model)
    if get_akida_version() == AkidaVersion.v1:
        _check_layer_inbounds(model)

    # Split model into theirs blocks:
    blocks = split_model_into_blocks(model)

    # This list will contains either a block converter instance,
    # or a list of non-skippable layers.
    straight_blocks = []
    # Evaluate block-by-block integrity
    for id, block in enumerate(blocks):
        # Initialize block_converter to None
        block_converter = None
        # Split blocks into skippable and none skippable blocks
        _, non_skippable = _extract_skippable_layers(block)
        # Skip the block if the block contains only skippable layers
        if len(non_skippable) == 0:
            continue
        # Get the corresponding BlockConverter of the layers block if available.
        # The first block is a special case and might target the HRC if fulfill some conditions
        if id == 0 and model.input_shape[-1] in (1, 3):
            block_converter = _get_input_block_converter(non_skippable)
        # If the first block doesn't match any Input pattern follow the classical way
        if not block_converter:
            block_converter = _get_block_converter(non_skippable)
        # One shouldn't get in here. If so the block pattern is unconvertible. Raise an error.
        if not block_converter:
            layers = [type(layer).__name__ for layer in non_skippable]
            supported_patterns = _display_supported_patterns(layers)
            raise RuntimeError(f"Invalid block pattern to conversion. Receives {layers}. "
                               f"Supported patterns:\n{supported_patterns}")
        straight_blocks.append(block_converter(non_skippable))

    return straight_blocks


def _check_model_input_output(model):
    """Asserts that model inputs/outputs are supported for conversion.

    The Keras model must have only one input layer and one output layer.
    On Akida 1.0, the input shape must 4-D (N, H, W, C).

    Args:
        model (tf.keras.model): the Keras model to check.
    """
    # Error if all model layers are skippable
    if all(isinstance(layer, (*skippable_layers, qlayers.Dequantizer)) for layer in model.layers):
        raise RuntimeError("Cannot convert model with only skippable layers.")

    # If the quantized keras model is Sequential, it should have its input_shape defined.
    if isinstance(model, Sequential):
        input_shape = getattr(model.layers[0], "input_shape", None)
        if input_shape is None:
            raise AttributeError(f"The Sequential quantized model {model.name} should have its"
                                 " input_shape fully defined.")

    # Error if multiple inputs
    if len(model.input_names) > 1:
        raise RuntimeError("Model must have only one input layer. Receives "
                           f"inputs {model.input_names}.")

    # Error if multiple outputs
    if len(model.output_names) != 1:
        raise RuntimeError("Model must have only one output layer. Receives"
                           f"outputs {model.output_names}.")

    # Error if input shape is not 2D or 4D
    if len(model.input_shape) not in (2, 4) and get_akida_version() == AkidaVersion.v1:
        raise RuntimeError(
            "Input shape of model must be 2-D or 4-D (batch size + 1-D or 3-D "
            f"tensors). Receives input shape {model.input_shape}.")

    # In akida HW one can realise skip connection directly with the main layer
    if len(model.layers[0].outbound_nodes) > 1:
        raise RuntimeError("The input model layer can only have one outbound node.")


def _extract_skippable_layers(block):
    """Split block into skippable and non skippable layers

    Args:
        block (tf.keras.Layer): block to split.

    Returns:
        tuple: list of skippable and non skippable layers
    """
    skippable, non_skippable = [], []
    for layer in block:
        if isinstance(layer, skippable_layers):
            skippable.append(layer)
        elif isinstance(layer, qlayers.QuantizedReshape):
            in_shape = layer.input_shape
            out_shape = layer.output_shape
            in_dims = [x for x in in_shape if x != 1]
            out_dims = [x for x in out_shape if x != 1]
            if in_dims != out_dims:
                non_skippable.append(layer)
            else:
                skippable.append(layer)
        else:
            non_skippable.append(layer)
    return skippable, non_skippable



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedLayerNormalization to Akida.
"""
from akida import MadNorm
import quantizeml.layers as qlayers
import numpy as np

from ..akida_versions import AkidaVersion
from .layer_utils import get_inbound_layers
from .weights import broadcast_and_set_variable
from .outputs import set_output_v2_variables, parse_output_bits, parse_post_op_buffer_bits
from .block_converter import BlockConverter, register_conversion_patterns
from .blocks import get_block_out_quantizer

__all__ = ["MadNormBlockConverter"]

_PATTERNS = [(qlayers.QuantizedLayerNormalization,)]


def _set_madnorm_variables(ak_layer, block):
    """Computes and sets the variables for an Akida MadNorm layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.
    """
    madnorm = block[0]
    assert isinstance(madnorm.gamma_quantizer, qlayers.WeightQuantizer)
    assert isinstance(madnorm.beta_quantizer, qlayers.AlignedWeightQuantizer)

    # get the QuantizedLayerNormalization gamma and shift
    gamma_ak = madnorm.gamma_quantizer.qweights.value.values.numpy()
    gamma_shift = madnorm.gamma_shift.value.numpy().astype(np.uint8)
    # get the QuantizedLayerNormalization beta and shift
    beta_quantizer = madnorm.beta_quantizer
    beta = beta_quantizer.qweights.value.values.numpy().astype(np.int32)
    beta_shift = beta_quantizer.shift.value.numpy().astype(np.uint8)
    beta_ak = (beta >> beta_shift).astype(np.int8)

    variables_ak = ak_layer.variables

    input_shift = getattr(madnorm, 'input_shift', None)
    if input_shift is not None:
        broadcast_and_set_variable(variables_ak, "input_shift",
                                   input_shift.value.numpy().astype(np.uint8))
    variables_ak["gamma"] = gamma_ak.astype(np.int8)
    broadcast_and_set_variable(variables_ak, "gamma_shift", gamma_shift)
    variables_ak["beta"] = beta_ak.astype(np.int8)
    broadcast_and_set_variable(variables_ak, "beta_shift", beta_shift)
    out_quantizer = get_block_out_quantizer(block)
    if out_quantizer:
        set_output_v2_variables(ak_layer, out_quantizer)


def _create_madnorm(block):
    """Parses a quantizeml QuantizedLayerNormalization layer and returns the
    params to create the corresponding Akida MadNorm layer.

    Args:
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        :obj:`akida.MadNorm`: The created akida layer.
    """
    madnorm = block[0]
    # In quantizeml one bit is reserved automatically for the sign, but in
    # akida this is rather checked during the clipping operations.
    block_params = {"buffer_bits": madnorm.buffer_bitwidth + 1}
    # parse the block output bits
    parse_output_bits(block, block_params)
    # parse the block post op buffer bits
    parse_post_op_buffer_bits(block, block_params)

    return MadNorm(**block_params,
                   name=madnorm.name)


def convert_quantized_madnorm(model_ak, block):
    """Converts QuantizedLayerNormalization layer and its variables and adds
    it to the Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        bool: Returns True for a successful conversion.
    """
    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, block[0])
    # Create and add layer to the akida model
    layer_ak = _create_madnorm(block)
    model_ak.add(layer_ak, inbound_layers_ak)
    # Set the akida layer converted variables
    _set_madnorm_variables(layer_ak, block)
    return True


class MadNormBlockConverter(BlockConverter):
    """Main class that should be used to check if the madnorm layer block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida v2 MadNorm layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._madnorm_additional_checks()

    def _madnorm_additional_checks(self):
        madnorm = self._block[0]
        # if input_shift available, make sure it has no negative values
        input_shift = getattr(madnorm, 'input_shift', None)
        if input_shift is not None:
            if input_shift.value and np.any(input_shift.value < 0):
                raise RuntimeError(
                    f"Layer {madnorm.name} contains negative values for " +
                    "input_shift, that is not supported")

    def convert(self, model_ak):
        return convert_quantized_madnorm(model_ak, self._block)


# Register the valid madnorm block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, MadNormBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert keras activation layers parameters and variables to akida.
"""
import numpy as np
from quantizeml.layers import QuantizedReLU, AlignedWeightQuantizer, OutputQuantizer
from .weights import broadcast_and_set_variable


def v1_relu_checks(layer):
    """Additionnal checks on QuantizedReLU layers of akida v1 models.

    Args:
        layer (:obj:`tf.keras.Layer`): the QuantizedReLU layer to check.
    """
    lname = layer.name
    if layer.max_value is None:
        raise ValueError(f"{lname}: in AkidaVersion.v1, unbounded QuantizedReLU is not supported.")
    assert isinstance(layer.max_value_quantizer, AlignedWeightQuantizer)
    # Check if there is an output_quantizer
    out_quantizer = getattr(layer, "out_quantizer", None)
    if not out_quantizer:
        raise ValueError(f"{lname}: in AkidaVersion.v1, output_quantizer is mandatory.")
    assert isinstance(out_quantizer, OutputQuantizer)
    if out_quantizer._axis == "per-axis":
        raise ValueError(f"{lname}: in AkidaVersion.v1, output_quantizer must be per-tensor.")


def parse_relu_v1(layer_k):
    """Parses the quantizeml.QuantizedReLU parameters for Akida v1 layers.

    Args:
        layer_k (:obj:`tf.keras.Layer`): the QuantizedReLU layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """
    # Check if there is an output_quantizer
    out_quantizer = getattr(layer_k, "out_quantizer", None)

    # In AkidaVersion.v1, output_quantizer is mandatory
    assert out_quantizer is not None

    return {'activation': True, 'act_bits': out_quantizer.bitwidth}


def parse_relu_v2(layer_k):
    """Parses the quantizeml.QuantizedReLU parameters for Akida v2 layers.

    Args:
        layer_k (:obj:`tf.keras.Layer`): the QuantizedReLU layer to parse.

    Returns:
        dict: the corresponding akida parameters.
    """

    return {'activation': True}


def set_relu_variables(layer_ak, layer_k):
    """Computes and sets the activation variables in an akida v2 layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida layer.
        layer_k (:obj:`quantizeml.QuantizedRelu`): the source QuantizedReLU layer.
    """

    assert isinstance(layer_k, QuantizedReLU)

    variables_ak = layer_ak.variables

    if layer_k.max_value is not None:
        max_value_quantizer = layer_k.max_value_quantizer
        assert isinstance(max_value_quantizer, AlignedWeightQuantizer)
        max_value = max_value_quantizer.qweights.value.values.numpy().astype(np.int32)
        max_value_shift = max_value_quantizer.shift.value.numpy().astype(np.uint8)
        max_value_ak = (max_value >> max_value_shift).astype(np.uint8)
        broadcast_and_set_variable(variables_ak, "max_value", max_value_ak)
        broadcast_and_set_variable(variables_ak, "max_value_shift", max_value_shift)



# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Implementation of the Adaround algorithm described in:
   Up or Down? Adaptive Rounding for Post-Training Quantization
   Markus Nagel, Rana Ali Amjad, Mart van Baalen, Christos Louizos, Tijmen
   Blankevoort
   https://arxiv.org/abs/2004.10568
"""
import shutil
from math import pi
import keras
import numpy as np
import tensorflow as tf

from ..transforms.clone import clone_model_with_weights
from ..quantization_layers import QuantizedDense, QuantizedConv2D, QuantizedSeparableConv2D
from ..quantization_ops import WeightQuantizer
from .calibration import QuantizationSampler, is_quantized_neural

# Rectified sigmoid constants
GAMMA = -0.1
ZETA = 1.1
# Rounding loss annealing constants
BETA_START = 20
BETA_END = 2
# Rounding loss regularization
RD_REG = 0.01


class AdaRoundQuantizer(WeightQuantizer):
    """Adaptive Rounding weights quantizer

    Replaces original weights by floored weights + decimal.

    It relies on an internal continuous variable to evaluate the decimals of
    each weights.

    It provides a rounding loss method that allows an optimizer to force the
    decimals to take the values 0 or 1.

    It also implements the WeightQuantizer API to be able to replace the
    original quantizers of a Model, and return the reconstructed weights
    instead of the quantized weights.

    Args:
        quantizer (:obj:`cnn2snn.quantization_ops.WeightQuantizer`): the
           original quantizer.
        weights (:obj:`np.ndarray`): the original weights
    """

    def __init__(self, quantizer, weights, **kwargs):
        # Store original quantizer
        self.quantizer = quantizer
        # Cache scale factor for the original weights
        self.sf = quantizer.scale_factor(weights)
        # Evaluate rescaled weights
        w_scaled = tf.constant(weights) * self.sf
        # Store floored rescaled weights
        self.w_floor = tf.constant(tf.math.floor(w_scaled))
        # Evaluate decimals
        d = w_scaled - self.w_floor
        # Create alpha variable, init so that self.decimals() == d
        initial_alpha = -1 * tf.math.log((ZETA - GAMMA) / (d - GAMMA) - 1)
        self.alpha = tf.Variable(name="alpha",
                                 initial_value=initial_alpha,
                                 dtype=tf.float32,
                                 trainable=True)
        # The quantizer behaviour is different at training or evaluation time
        self.training = True
        super().__init__(quantizer.bitwidth, **kwargs)

    def decimals(self):
        """The projection of the alpha variable to [0, 1]

        Returns:
            :obj:`tf.Tensor` : the reconstructed decimals
        """
        # Evaluate sigmoid(alpha) to evaluate the rounding between [0, 1]
        rounding = tf.math.sigmoid(self.alpha)
        # Apply an affine transformation to project to [GAMMA, ZETA]
        rounding = rounding * (ZETA - GAMMA) + GAMMA
        # Clip and return
        return tf.clip_by_value(rounding, 0, 1)

    def loss(self, beta):
        """The rounding loss used to force the decimals to 0 or 1.

        Args:
            beta (int): the loss annealing parameter

        Returns:
            :obj:`tf.Tensor`: the rounding loss Tensor
        """
        if beta == 0:
            return 0
        # The loss is the sum of the distances to the [0, 1] boundaries powered
        # by the beta annealing parameter
        return tf.reduce_sum(1 - (tf.abs(2 * self.decimals() - 1))**beta)

    def status(self):
        """The current rounding status

        This primarily evaluates the number of roamers, i.e. weights that have
        not be rounded yet.
        This also returns the number of floored and ceiled weights.

        Returns:
            (int, int, int): the number of roamers, floored and ceiled weights
        """
        d = self.decimals().numpy()
        num_weights = d.size
        roamers = num_weights - np.sum(np.isin(d, [0, 1]))
        floor_r = np.sum(d <= 0.5) / num_weights
        ceil_r = np.sum(d > 0.5) / num_weights
        return roamers, floor_r, ceil_r

    def recons_weights(self):
        """Reconstruct weights from original weights using current decimals.

        Returns:
            :obj:`tensorflow.Tensor`: a Tensor of float weights.
        """
        # Reconstruct rescaled weights: floored rescaled weights + decimals
        adaround_w = self.w_floor + self.decimals()
        # Clip and rescale to obtain the rounded weights
        return tf.clip_by_value(adaround_w, -self.kmax_, self.kmax_) / self.sf

    def scale_factor(self, _w):
        """Return the original quantizer scale factor

        Args:
          _w (:obj:`tensorflow.Tensor`): ignored.

        Returns:
          :obj:`tensorflow.Tensor`: a Tensor containing a list of scalar values
                (1 or more).
        """
        return self.sf

    def quantize(self, _w):
        """Return layer weights

        Although it takes the original weights as parameters, it ignores it and
        uses instead the reconstructed weights.

        This method has a different behaviour at evaluation and training time.

        - at training time it simply returns the non-quantized reconstructed
          weights to provide a smooth quantization loss
        - at evaluation time it applies the original quantization to the
          reconstructed weights.

        Note that when all decimals have converged to either zero or one, the
        reconstructed weights are aligned on quantization intervals, so the
        returned weights are equivalent to the reconstructed weights.

        Args:
            _w (:obj:`tensorflow.Tensor`): ignored.

        Returns:
            :obj:`tensorflow.Tensor`: a Tensor of quantized weights.
        """
        if self.training:
            return self.recons_weights()
        return self.quantizer.quantize(self.recons_weights())


def initialize_training(optimizer, model):
    """Prepare target model for Adaround optimization

    The original quantizers are replaced by AdaRound quantizers

    Args:
        optimizer (:obj:`tensorflow.keras.optimizers.Optimizer`): an optimizer
        model (:obj:`keras.models.Model`): a quantized Keras Model

    Returns:
        dict of [`AdaRoundQuantizer`], dict of [`tf.Variable`]: a list of AdaRound quantizers and
        a list of trainable variables, grouped by layer

    """
    adarounds = {}
    trainable_variables = {}
    for layer in model.layers:
        if not is_quantized_neural(layer):
            continue
        # Replace standard quantizers by Adaround quantizers
        adarounds[layer] = []
        trainable_variables[layer] = []
        if isinstance(layer, QuantizedSeparableConv2D):
            w_list = layer.get_weights()
            dw = w_list[0]
            layer.quantizer_dw = AdaRoundQuantizer(layer.quantizer_dw, dw)
            adarounds[layer].append(layer.quantizer_dw)
            trainable_variables[layer].append(layer.quantizer_dw.alpha)
            pw = w_list[1]
            layer.quantizer = AdaRoundQuantizer(layer.quantizer, pw)
            adarounds[layer].append(layer.quantizer)
            trainable_variables[layer].append(layer.quantizer.alpha)
        elif isinstance(layer, (QuantizedDense, QuantizedConv2D)):
            w_list = layer.get_weights()
            w = w_list[0]
            layer.quantizer = AdaRoundQuantizer(layer.quantizer, w)
            adarounds[layer].append(layer.quantizer)
            trainable_variables[layer].append(layer.quantizer.alpha)
    # Create all optimizer variables
    all_trainable_variables = sum(trainable_variables.values(), [])
    # Since TensorFlow 2.12, the optimizer must build the variables before use them.
    if hasattr(optimizer, 'build'):
        optimizer.build(all_trainable_variables)
    return adarounds, trainable_variables


def finalize_training(layer):
    """Restore target layer after Adaround optimization

    The original quantizers are restored and the rounded weights are frozen.

    Args:
        layer (:obj:`keras.layers.Layer`): the layer to optimize

    """
    # Freeze rounded weights and restore original quantizers
    if not is_quantized_neural(layer) or not isinstance(layer.quantizer,
                                                        AdaRoundQuantizer):
        return
    if isinstance(layer, QuantizedSeparableConv2D):
        # Get new rounded weights
        adaround_dw = layer.quantizer_dw.recons_weights()
        adaround_pw = layer.quantizer.recons_weights()
        # Restore original quantizers
        layer.quantizer_dw = layer.quantizer_dw.quantizer
        layer.quantizer = layer.quantizer.quantizer
        # Compare quantized weights: original vs rounded
        w_list = layer.get_weights()
        print_console(
            "Changed dw weights: {} / {}".format(
                np.sum(
                    layer.quantizer_dw.quantize(
                        w_list[0]) != layer.quantizer_dw.quantize(adaround_dw)),
                adaround_dw.numpy().size))
        print_console("Changed pw weights: {} / {}".format(
            np.sum(
                layer.quantizer.quantize(w_list[1]) != layer.quantizer.quantize(
                    adaround_pw)),
            adaround_pw.numpy().size))
        # Replace original weights by rounded weights
        w_list[0] = adaround_dw
        w_list[1] = adaround_pw
        layer.set_weights(w_list)
    elif isinstance(layer, (QuantizedConv2D, QuantizedDense)):
        # Get new rounded weights
        adaround_w = layer.quantizer.recons_weights()
        # Restore original quantizer
        layer.quantizer = layer.quantizer.quantizer
        # Compare quantized weights: original vs rounded
        w_list = layer.get_weights()
        print("Changed weights: {} / {}".format(
            np.sum(
                layer.quantizer.quantize(w_list[0]) != layer.quantizer.quantize(
                    adaround_w)),
            adaround_w.numpy().size))
        # Replace original weights by rounded weights
        w_list[0] = adaround_w
        layer.set_weights(w_list)


def compute_beta(step, max_step):
    """Computes the rounding loss annealing parameter

    Args:
        step (int): the current optimization step
        max_step (int): the maximum optimization step

    Returns:
        int: the annealing parameter
    """
    rel_step = step / max_step
    cosine_decay = 0.5 * (1 + tf.math.cos(rel_step * pi))
    return BETA_END + (BETA_START - BETA_END) * cosine_decay


def print_console(message, new_line=True):
    """Print a message on the console.

    If new_line is True, this is equivalent to a simple print.

    If new_line is False, the message is padded with spaces until the end of
    the line, but no end-of-line is inserted, which means that subsequent calls
    to print will overwrite the last line of console.

    Args:
        message (str): the message to print on the console
        new_line (bool): go to the next line after the message has been printed
    """
    if new_line:
        print(message)
    else:
        col, _ = shutil.get_terminal_size()
        print(f"{message:<{col}}", sep='', end='\r', flush=True)


def optimize_rounding(model, samples, layer, optimizer, epochs, loss,
                      batch_size, include_activation, adarounds, trainable_variables):
    """AdaRound main optimization method

    Optimize the AdaRound alpha variables using gradient descent to minimize
    the quantization error.

    Args:
        model (:obj:`keras.models.Model`): a quantized Keras Model
        samples (:obj:`np.ndarray`): a set of samples used for calibration
        layer (:obj:`keras.layers.Layer`): the layer to optimize
        optimizer (:obj:`tensorflow.keras.optimizers.Optimizer`): an optimizer
        epochs (int): the maximum number of epochs
        loss (:obj:`tensorflow.keras.losses.Loss`): the error loss function
        batch_size (int): the batch size used when evaluating samples
        include_activation (bool): quantization error is evaluated after
          activation.
    """
    # Instantiate a sampler and select the target layer
    sampler = QuantizationSampler(model, samples, batch_size)
    sampler.select_layer(layer, include_activation)
    m = keras.metrics.MeanSquaredError()
    # Evaluate quantization error
    err_before = sampler.quantization_error(m)
    float_outputs = None
    # Training loop
    print_console("Optimizing {}:".format(layer.name))
    for e in range(epochs):
        # Evaluate rounding loss parameter
        beta = compute_beta(e, epochs)
        for i in range(sampler.n_batches):

            # If we have a single batch, we can cache the float outputs
            if sampler.n_batches > 1 or float_outputs is None:
                # Evaluate the float outputs for this batch
                float_outputs = sampler.float_outputs(i)

            # Optimize the adaround alpha variables
            with tf.GradientTape() as tape:

                # Compute layer outputs with adaround
                outputs = sampler.quantized_outputs(i)

                # Evaluate quantization loss
                q_loss = loss(float_outputs, outputs)

                # Evaluate rounding loss for each set of weights
                r_loss = 0
                for adr in adarounds:
                    r_loss += RD_REG * adr.loss(beta)

                total_loss = q_loss + r_loss

            # Get gradients of loss wrt alpha
            gradients = tape.gradient(total_loss, trainable_variables)

            # Update the alpha variables
            optimizer.apply_gradients(zip(gradients, trainable_variables))

        log = f"Epoch {e + 1}"
        log += f" - loss (quantization|rounding) {q_loss:.8f} | {r_loss:.8f}"
        for adr in adarounds:
            roamers, floor_r, ceil_r = adr.status()
            log += f" - roamers: {roamers} - ↓↑ [{floor_r:.2f}, {ceil_r:.2f}]"
        print_console(log, new_line=False)
        if beta > 0 and r_loss == 0:
            # All weights have been rounded
            break
    # Go to the next line in console
    print_console("")
    # Set Adaround quantizer to evaluation mode
    for adr in adarounds:
        adr.training = False
    # Evaluate final quantization error
    err_after = sampler.quantization_error(m)
    print_console(f"quantization error: {err_before:.4f} -> {err_after:.4f}")
    # Freeze optimized weights
    finalize_training(layer)


def adaround(model,
             samples,
             optimizer,
             epochs,
             loss=tf.keras.losses.MeanSquaredError(),
             batch_size=None,
             include_activation=False):
    """Optimize the rounding of quantized weights.

    This implements the Adaround algorithm described in:
    Up or Down? Adaptive Rounding for Post-Training Quantization
    Markus Nagel, Rana Ali Amjad, Mart van Baalen, Christos Louizos, Tijmen
    Blankevoort
    https://arxiv.org/abs/2004.10568

    Instead of rounding weights to the nearest, Adaround introduces a tensor of
    continuous variables representing the decimals of the float weights, and
    thus formulates the minimization of the quantization error as a  Quadratic
    Unconstrained Binary Optimization problem, iteratively pushing the decimal
    variables to a distribution of 0 and 1 minimizing the error.

    After the optimization, the quantization scales are preserved, but each
    weight is closer or equal to a quantized value.

    When optimizing a model, the following must be provided:

    - a set of samples (typically from the training dataset),
    - an optimizer,
    - the maximum number of epochs (the optimization of a layer stops when all
      weights have been rounded).

    Args:
        model (:obj:`keras.Model`): a quantized Keras Model
        samples (:obj:`np.ndarray`): a set of samples used for calibration
        optimizer (:obj:`tensorflow.keras.optimizers.Optimizer`): an optimizer
        epochs (int): the maximum number of epochs
        loss (:obj:`tensorflow.keras.losses.Loss`): the error loss function
        batch_size (int): the batch size used when evaluating samples
        include_activation (bool): quantization error is evaluated after
          activation.

    Returns:
        :obj:`tf.Model`: a quantized Keras model whose weights have been
        optimized
    """
    adaround_model = clone_model_with_weights(model)
    # Initialize all trainable variables
    adarounds, trainable_variables = initialize_training(optimizer, adaround_model)

    for layer in adaround_model.layers:
        if is_quantized_neural(layer):
            # Optimize rounding
            optimize_rounding(
                adaround_model,
                samples,
                layer,
                optimizer,
                epochs,
                loss,
                batch_size,
                include_activation,
                adarounds[layer],
                trainable_variables[layer])
    return adaround_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["set_output_scale_variables"]

import numpy as np

from .weights import broadcast_and_set_variable


def set_output_scale_variables(ak_layer, scale=1, shift=0):
    """Set output scale into akida variables.

    Args:
        ak_layer (akida.Layer): the akida layer to set variables.
        scale (np.ndarray): the scale. Defaults to 1.
        shift (np.ndarray): the power of two that represent the shift. Defaults to 0.
    """
    if "output_scales" in ak_layer.variables.names:
        if np.any(scale <= 0):
            raise ValueError(f"Error found when setting scale in {ak_layer.name}:",
                             f"{scale} must be higher than zero.")
        broadcast_and_set_variable(ak_layer.variables, "output_scales", scale)

    # Shift integrity check
    ak_shift = np.round(np.log2(shift))
    if np.any(2.0**ak_shift != shift) or np.any(shift == 0):
        raise ValueError(f"Error found when setting shift in {ak_layer.name}:",
                         f"{shift} is not a power of two.")

    # In akida, shift is applied as left shift (when positive) or right shift (otherwise).
    # However, in onnx scale out, shift was performed through one division.
    # It explains the minus sign.
    broadcast_and_set_variable(ak_layer.variables, "output_shift", -ak_shift)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedConcatenate to Akida.
"""
from akida import Concatenate
from quantizeml.layers import QuantizedConcatenate

from .layer_utils import get_inbound_layers
from .block_converter import BlockConverter, register_conversion_patterns
from ..akida_versions import AkidaVersion

__all__ = ["ConcatenateBlockConverter"]

_PATTERNS = [(QuantizedConcatenate,)]


def convert_quantized_concatenate(model_ak, block):
    """Converts QuantizedConcatenate layer and its variables and adds it to the
    Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.

    Returns:
        bool: Returns True for a successful conversion.
    """

    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, block[0])
    # Create and add layer to the akida model
    # It is quite simple because there are no variables or params to be set.
    layer_ak = Concatenate()
    model_ak.add(layer_ak, inbound_layers_ak)
    return True


class ConcatenateBlockConverter(BlockConverter):
    """Main class that should be used to check if the concatenate layer block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida Concatenate
    layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def convert(self, model_ak):
        return convert_quantized_concatenate(model_ak, self._block)


# Register the valid concatenate block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, ConcatenateBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedBatchNormalization to Akida.
"""
from akida import BatchNormalization
from quantizeml.layers import (QuantizedBatchNormalization, QuantizedReLU,
                               WeightQuantizer, AlignedWeightQuantizer)
import numpy as np

from .weights import broadcast_and_set_variable
from .outputs import set_output_v2_variables, parse_output_bits, parse_post_op_buffer_bits
from .activations import set_relu_variables
from .layer_utils import get_inbound_layers
from .block_converter import BlockConverter, register_conversion_patterns
from ..akida_versions import AkidaVersion
from .conv_common import get_layer_by_type
from .blocks import get_block_out_quantizer

__all__ = ["BatchNormBlockConverter"]

_PATTERNS = [(QuantizedBatchNormalization,),
             (QuantizedBatchNormalization, QuantizedReLU)]


def _set_batchnorm_variables(ak_layer, block):
    """Computes and sets the variables for an Akida BatchNormalization layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.
    """
    layer_bn = block[0]
    assert isinstance(layer_bn.a_quantizer, WeightQuantizer)
    assert isinstance(layer_bn.b_quantizer, AlignedWeightQuantizer)

    # get the QuantizedBatchNormalization a, b and shift
    a_ak = layer_bn.a_quantizer.qweights.value.values.numpy()
    b_quantizer = layer_bn.b_quantizer
    b = b_quantizer.qweights.value.values.numpy().astype(np.int32)
    b_shift = b_quantizer.shift.value.numpy().astype(np.uint8)
    b_ak = (b >> b_shift).astype(np.int8)

    variables_ak = ak_layer.variables

    input_shift = layer_bn.input_shift.value
    if input_shift is not None:
        broadcast_and_set_variable(variables_ak, "input_shift",
                                   input_shift.numpy().astype(np.uint8))

    variables_ak["a"] = a_ak.astype(np.int8)
    variables_ak["b"] = b_ak
    broadcast_and_set_variable(variables_ak, "b_shift", b_shift)

    relu = get_layer_by_type(block, QuantizedReLU)
    if relu:
        set_relu_variables(ak_layer, relu)

    out_quantizer = get_block_out_quantizer(block)
    if out_quantizer:
        set_output_v2_variables(ak_layer, out_quantizer)


def _create_batchnorm(block):
    """Parses a quantizeml QuantizedBatchNormalization layers block and returns the
    corresponding Akida v2 BatchNormalization layer.

    Args:
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        :obj:`akida.BatchNormalization`: The created akida layer.
    """
    bn = block[0]
    bn_params = {}
    relu = get_layer_by_type(block, QuantizedReLU)
    if relu:
        bn_params['activation'] = True
    # In quantizeml one bit is reserved for the sign in the buffer bitwidth
    # variable, but in akida this value has to be added back to have the
    # correct clipping.
    bn_params['buffer_bits'] = bn.buffer_bitwidth + 1

    # parse the block output bits
    parse_output_bits(block, bn_params)
    # parse the block post op buffer bits
    parse_post_op_buffer_bits(block, bn_params)

    return BatchNormalization(**bn_params,
                              name=bn.name)


def convert_batchnorm_block(model_ak, block):
    """Converts QuantizedBatchNormalization layer and its variables and adds
    it to the Akida's model. If followed by QuantizedReLU, it will set its
    variables too.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be
            added.
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        bool: Returns True for a successful conversion.
    """
    layer_ak = _create_batchnorm(block)
    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, block[0])
    # Add the layer to the Akida model
    model_ak.add(layer_ak, inbound_layers_ak)
    # Set its variables
    _set_batchnorm_variables(layer_ak, block)
    return True


class BatchNormBlockConverter(BlockConverter):
    """Main class that should be used to check if the batch norm layer block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida
    BatchNormalization layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def convert(self, model_ak):
        return convert_batchnorm_block(model_ak, self._block)


# Register the valid batch norm block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, BatchNormBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedBufferTempConv to Akida.
"""
from quantizeml.layers import QuantizedBufferTempConv, QuantizedReLU
from akida import BufferTempConv
from ..akida_versions import AkidaVersion
from .btc_common import parse_buf_temp_conv_block, set_buffer_temp_conv_block_variables
from .block_converter import BlockConverter, register_conversion_patterns

__all__ = ["BufferTempConvBlockConverter"]

_PATTERNS = [(QuantizedBufferTempConv,), (QuantizedBufferTempConv, QuantizedReLU)]


def _convert_buf_temp_conv_block(model_ak, block):
    """Converts a buffer temp conv block into an akida BufferTempConv layer.

    The expected sequence is:

    - QuantizedBufferTempConv,
    - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        block (list(:obj:`tf.keras.Layer`)): the block layers.

    Returns:
        bool: Returns True for a successful conversion.
    """

    # Evaluate the buffer temporal convolution layer parameters
    btc_layer_params = parse_buf_temp_conv_block(block)

    # Create Akida layer
    btc_layer_ak = BufferTempConv(**btc_layer_params)
    # Add layer to the model to build its internal variables
    model_ak.add(btc_layer_ak)

    # Set base variables
    set_buffer_temp_conv_block_variables(btc_layer_ak, block)

    return True


class BufferTempConvBlockConverter(BlockConverter):
    """Main class that should be used to check if the buffer temp conv block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida BufferTempConv
    layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._buffer_temp_conv_additional_checks()

    def _buffer_temp_conv_additional_checks(self):
        btc_layer = self._block[0]

        # weight bitwidth should be 8bit maximum
        weight_bits = btc_layer.weight_quantizer.bitwidth
        assert weight_bits <= 8, ("BufferTempConv layer handles weights with"
                                  f" maximum 8 bits. Received: {weight_bits}.")

    def convert(self, model_ak):
        return _convert_buf_temp_conv_block(model_ak, self._block)


# Register the valid buffer temp conv block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, BufferTempConvBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["Conv2DOnnxConverter", "InputConv2DOnnxConverter"]

import akida

from .base_converter import OnnxConverter
from .register import register_onnx_converter_target
from .conv_commons import check_convolution_compatibility, set_convolutional_variables
from .padding import get_akida_padding


def _get_conv_params(layer):
    # Parse common information
    layer_params = {
        "name": layer.name,
        "activation": layer.activation,
        "output_bits": 8,
        "filters": layer.weights["W"].shape[0],
        "kernel_size": layer.weights["W"].shape[-1],
        "padding": get_akida_padding(layer),
        "pool_type": layer.pool_type,
        "kernel_stride": layer.strides[0],
    }
    if layer.pool_type == akida.PoolType.Max:
        layer_params["pool_size"] = layer.pool_size[0]
        layer_params["pool_stride"] = layer.pool_strides[0]
    return layer_params


@register_onnx_converter_target("QuantizedConv2D")
class Conv2DOnnxConverter(OnnxConverter):
    """Convert QuantizedConv2D type node into an akida.Conv2D.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """

    def load_attributes(self, node):
        # Load default attributes
        super().load_attributes(node)

        # Some attributes should infer from node.op_type
        n_op = node.op_type
        self.activation = "ReLU" in n_op
        if "GlobalAvgPool" in n_op:
            self.pool_type = akida.PoolType.Average
        elif "MaxPool" in n_op:
            self.pool_type = akida.PoolType.Max
        else:
            self.pool_type = akida.PoolType.NoPooling
        self.use_bias = "Biased" in n_op

        # Padding type is inferred from pads attribute
        self.pads = self.weights["pads"].tolist()

    def _additional_checks(self):
        check_convolution_compatibility(self)

    def _parse_akida_layer(self):
        layer_params = _get_conv_params(self)
        return akida.Conv2D(**layer_params)

    def _set_akida_variables(self, ak_layer):
        set_convolutional_variables(self, ak_layer)


@register_onnx_converter_target("QuantizedInputConv2D")
class InputConv2DOnnxConverter(Conv2DOnnxConverter):
    """Convert QuantizedInputConv2D type node into an akida.InputConv2D.

    Args:
        node (NodeProto): the node to convert.
        model (ModelProto): the model that the node is.
    """
    @property
    def is_input_layer(self):
        return True

    def _additional_checks(self):
        check_convolution_compatibility(self, use_pad_value=True)

    def _parse_akida_layer(self):
        layer_params = _get_conv_params(self)
        return akida.InputConv2D(input_shape=self.input_shape, **layer_params)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedConv2DTranspose to Akida.
"""

from quantizeml.layers import (QuantizedConv2DTranspose, QuantizedReLU)
from akida import Conv2DTranspose
from ..akida_versions import AkidaVersion
from .block_converter import BlockConverter, register_conversion_patterns
from .conv_common import set_conv2d_transpose_block_variables, parse_conv2d_transpose_block

__all__ = ["ConvTransposeBlockConverter"]

_PATTERNS = [(QuantizedConv2DTranspose,), (QuantizedConv2DTranspose, QuantizedReLU)]


def convert_conv_transpose_block(model_ak, block):
    """Converts a convolutional transpose block into an akida Conv2DTranspose layer.

    The expected sequence is:

    - QuantizedConv2DTranspose,
    - QuantizedReLU (optional).

    Args:
        model_ak (:obj:`akida.Model`): the target Akida model.
        block (list(:obj:`tf.keras.Layer`)): the block layers.

    Returns:
        bool: Returns True for a successful conversion.
    """

    # Evaluate the convolutional transpose layer parameters
    conv_transpose_params = parse_conv2d_transpose_block(block)

    # Create Akida layer
    conv_transpose_ak = Conv2DTranspose(**conv_transpose_params)
    # Add layer to the model to build its internal variables
    model_ak.add(conv_transpose_ak)

    # Set base variables
    set_conv2d_transpose_block_variables(conv_transpose_ak, block)

    return True


class ConvTransposeBlockConverter(BlockConverter):
    """Main class that should be used to check if the conv transpose block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida Conv2DTranspose
    layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._conv_transpose_additional_checks()

    def _conv_transpose_additional_checks(self):
        conv_transpose = self._block[0]

        assert conv_transpose.kernel_size[0] == conv_transpose.kernel_size[1], (
            "QuantizedConv2DTranspose kernel should be square")
        assert conv_transpose.kernel_size[0] in (3, 4)
        assert conv_transpose.strides[0] == conv_transpose.strides[1], (
            "QuantizedConv2DTranspose strides should be the same on both dimensions.")

        assert conv_transpose.strides[0] == 2, ("QuantizedConv2DTranspose handles only stride 2")
        # Padding value must be built in constructor
        assert conv_transpose.padding == "same", ("QuantizedConv2DTranspose handles only 'same' "
                                                  "padding")
        # The only weight bitwidth supported is [4, 8]
        weight_bits = conv_transpose.weight_quantizer.bitwidth
        assert weight_bits in [4, 8], ("QuantizedConv2DTranspose layer can only handle weights"
                                       f" with 4 or 8 bits. Received: {weight_bits}.")

    def convert(self, model_ak):
        return convert_conv_transpose_block(model_ak, self._block)


# Register the valid conv2d transpose block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, ConvTransposeBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["set_activation_variables"]

from .weights import broadcast_and_set_variable, to_value_shift


def set_activation_variables(ak_layer, max_value=None):
    """Set max value into akida variables.

    Args:
        ak_layer (akida.Layer): the akida layer to set variables.
        max_value (int, optional): the maximal value to set. Defaults to None.
    """
    if max_value is not None:
        # Max value is converted to a fixed point: unsigned 8bit << unsigned 8bit
        max_value, max_value_shift = to_value_shift(max_value, signed=False)
    else:
        # When max_value is not provided, resolve threshold as maximal SHIFTED int32 value
        max_value, max_value_shift = 1, 30
    broadcast_and_set_variable(ak_layer.variables, "max_value", max_value)
    broadcast_and_set_variable(ak_layer.variables, "max_value_shift", max_value_shift)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Utility function to convert block layers to akida.
"""
import warnings
from quantizeml.layers import (QuantizedReLU,
                               QuantizedExtractToken,
                               QuantizedShiftmax,
                               Dequantizer)
from ..akida_versions import AkidaVersion, get_akida_version


def split_model_into_blocks(model):
    """Search into the model the sets of possible blocks and return them.

    A set of layers are considered as block if:
        1. There is only one output quantizer (in the last layer of the block)
        2. There are no multi-inbound layers
        3. There are no QuantizedExtractToken layers

    Layers with more than one inbound layer will be handled separately.

    Args:
        model (:obj:`tf.models.Model`): The model to split

    Returns:
        list: a list of sequences of layers ('blocks').
    """
    def _get_inbound_layers(target_layer):
        inbound = target_layer.inbound_nodes[0].inbound_layers
        return inbound if isinstance(inbound, (list, tuple)) else [inbound]

    def _check_single_layer_block(target_layer):
        # Single blocks will be multi-inputs and QuantizedExtractToken (with axis==1)
        prev_layers = _get_inbound_layers(target_layer)
        if len(prev_layers) > 1:
            return True
        if (isinstance(target_layer, QuantizedExtractToken) and target_layer.axis == 1):
            return True
        return False

    def _search_block_v1(target_layer):
        # We consider the target is part of one block if there are previous
        # layers. (Akida v1 doesn't support multi inbound layers)
        prev_layers = _get_inbound_layers(target_layer)
        # If it's the first layer return it
        if len(prev_layers) == 0:
            return [target_layer]

        # If previous layer has an output quantizer, the blocks end here.
        next_target = prev_layers[0]
        out_quantizer = getattr(next_target, "out_quantizer", None)
        if out_quantizer:
            if not isinstance(next_target, QuantizedReLU):
                raise RuntimeError("Model incompatible with akida v1 version. The layer"
                                   f" {next_target.name} followed by {target_layer.name}"
                                   " is not valid.")
            return [target_layer]

        # Otherwise, we continue searching...
        return _search_block_v1(next_target) + [target_layer]

    def _search_block_v2(target_layer):
        # We consider the target is part of one block if there are previous
        # layers (target is not the model's input), or it's multi-inbound.
        prev_layers = _get_inbound_layers(target_layer)
        if len(prev_layers) != 1:
            return [target_layer]
        # If previous layer has an output quantizer, the blocks end here.
        next_target = prev_layers[0]
        out_quantizer = getattr(next_target, "out_quantizer", None)
        # Note that the out_quantizer marks the limit of a block only if it has a
        # bitwidth <9 or if bitwidth > 9 and followed by a QuantizedShiftmax
        if (out_quantizer and (out_quantizer.bitwidth < 9 or (out_quantizer.bitwidth >= 9
                                                              and isinstance(target_layer,
                                                                             QuantizedShiftmax)))
                or _check_single_layer_block(next_target)):
            return [target_layer]

        # Otherwise, we continue searching...
        return _search_block_v2(next_target) + [target_layer]

    def _search_block(layer):
        return _search_block_v1(layer) if get_akida_version() == AkidaVersion.v1 \
            else _search_block_v2(layer)

    def _get_deq_id(model):
        """Helper to get the first dequantizer id.

        Args:
            model (:obj:`tf.models.Model`): the model to analyse.

        Returns:
            int: the first dequantized id or -1 if not found.
        """
        for layer_id, layer in enumerate(model.layers):
            if isinstance(layer, Dequantizer):
                return layer_id

        return -1

    def _get_layers_before_dequantizer(model):
        # Split layers on the first dequantizer layer
        deq_id = _get_deq_id(model)
        if deq_id >= 0:
            layers_pred = model.layers[:deq_id]
            layers_next = model.layers[deq_id + 1:]
            deq_layer = model.layers[deq_id]
        else:
            layers_pred = model.layers
            layers_next = []
            deq_layer = None

        return deq_layer, layers_pred, layers_next

    def _is_end_of_block(layer, blocks):
        # Return true if the layer is the end of a new block.
        # The very last layer is the end of another block
        if not blocks:
            return True
        outbounds = layer.outbound_nodes
        # If there are many outbounds, this is an end of block
        if len(outbounds) > 1:
            return True
        # Prepare a list of all first layers of all blocks
        first_layers = [block[0] for block in blocks]
        outbound = outbounds[0].outbound_layer
        return outbound in first_layers

    blocks = []
    # If there is a Dequantizer, save the first one as single block layer and then
    # remove all layers after it (not supported at conversion time)
    deq_layer, model_layers, skip_layers = _get_layers_before_dequantizer(model)
    if deq_layer:
        blocks.append([deq_layer])

    # Prints a warning message with a summary of the skipped layers
    if len(skip_layers) > 0:
        stop_layer_msg = " at layer " + model_layers[-1].name if len(model_layers) > 0 else ""
        skip_layers_summary = "___________________________________________________\n"
        skip_layers_summary += "Layer (type)\n"
        skip_layers_summary += "===================================================\n"
        for layer in skip_layers:
            skip_layers_summary += layer.name + " (" + layer.__class__.__name__ + ")\n"
        skip_layers_summary += "===================================================\n"
        warnings.warn("Conversion stops" + stop_layer_msg + " because of a dequantizer. "
                      "The end of the model is ignored:\n" + skip_layers_summary)

    # We forward from head to bottom
    for layer in model_layers[::-1]:
        if _is_end_of_block(layer, blocks):
            new_block = _search_block(layer)
            blocks.append(new_block)
        # The skip layers must be inside of one block
        elif layer not in sum(blocks, []):
            raise RuntimeError(f"Impossible to append {layer.name} into a block.")

    # At the end, we sort blocks from bottom to head
    return blocks[::-1]


def get_block_out_quantizer(block):
    """Extract the output_quantizer of the layers block.

    Note that a block of layers is considered as block if:
        1. There is only one output quantizer (in the last layer of the block)
        2. There are no multi-inbound layers
        3. There are no QuantizedExtractToken layers

    This method should ideally be called after "split_model_into_blocks" to make sure
    that the layers block can be handled correctly.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the layers block.

    Returns:
        :obj:`quantizeml.layers.OutputQuantizer`: the layers block output quantizer (None if the
            block has no output quantizer)
    """
    # Output quantizer is on the last layer of the block
    return getattr(block[-1], "out_quantizer", None)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedExtractToken to Akida.
"""
from akida import ExtractToken
import quantizeml.layers as qlayers

from .layer_utils import get_inbound_layers
from .block_converter import BlockConverter, register_conversion_patterns
from ..akida_versions import AkidaVersion


__all__ = ["ExtractTokenBlockConverter"]

_PATTERNS = [(qlayers.QuantizedExtractToken,)]


def _get_layer_token_range(layer):
    """Helper to extract the layer token range.

    Args:
        layer (:obj:`tf.keras.Layer`): the quantizeml QuantizedExtractToken layer.

    Returns:
        tuple, list: token range of the layer in the following format ((begin, end), token_range).
    """
    assert isinstance(layer, qlayers.QuantizedExtractToken)
    # Akida is capable of supporting only a given combination of token
    if isinstance(layer.token, int):
        token_range = [layer.token]
    else:
        token_range = [*layer.token]
    begin = min(token_range)
    end = max(token_range) + 1

    return (begin, end), token_range


def _create_extract_token(block):
    """Parses a quantizeml QuantizedExtractToken layer and returns the
    params to create the corresponding Akida v2 ExtractToken layer.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.

    Returns:
        :obj:`akida.ExtractToken`: The created akida layer.
    """
    (begin, end), _ = _get_layer_token_range(block[0])

    extract_token = ExtractToken(begin=begin,
                                 end=end,
                                 name=block[0].name)
    return extract_token


def convert_extract_token(model_ak, block):
    """Converts QuantizedExtractToken layer and adds it to the Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be
            added.
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.

    Returns:
        bool: Returns True for a successful conversion.
    """

    # Retrieve the akida inbound layers
    inbound_layers_ak = get_inbound_layers(model_ak, block[0])
    # Create and add layer to the akida model
    # It is quite simple because there are no variables or params to be set.
    layer_ak = _create_extract_token(block)
    model_ak.add(layer_ak, inbound_layers_ak)
    return True


class ExtractTokenBlockConverter(BlockConverter):
    """Main class that should be used to check if the extract token layer block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida ExtractToken
    layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def __init__(self, block):
        super().__init__(block)
        self._extract_token_additional_checks()

    def _extract_token_additional_checks(self):
        layer = self._block[0]
        (begin, end), token_range = _get_layer_token_range(layer)
        # Check range is continuous
        continuous = sorted(token_range) == list(range(begin, end))
        if not continuous:
            raise ValueError(f"Argument token in {layer.name} should contain a "
                             "continuous range")

    def convert(self, model_ak):
        return convert_extract_token(model_ak, self._block)


# Register the valid extract token block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, ExtractTokenBlockConverter)



# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Command-line interface"""

import argparse
import re
import os

import tensorflow as tf
import numpy as np
import onnx

from .converter import convert
from .quantization import quantize
from .utils import load_quantized_model
from .transforms import reshape, normalize_separable_model
from .calibration import adaround, bias_correction, activations_rescaling


def quantize_model(model_path, wq, aq, iq, fold_BN):
    """Wrapper to quantize model"""
    model = load_quantized_model(model_path)
    if iq == -1:
        iq = wq
    model_q = quantize(model, wq, aq, iq, fold_BN)
    # Extract base name
    base_name = os.path.splitext(model_path)[0]
    # Cross-platform path may contain alpha characters, /, \ and :
    path_re = r"([\w/\\:~]+)"
    # Quantization suffix has a well-identified structure
    suffix_re = r"(_iq\d_wq\d_aq\d)"
    p = re.compile(path_re + suffix_re)
    # Look for an existing quantization suffix in the base name
    m = p.match(base_name)
    if m:
        # Only keep the actual base name (group(2) contains the suffix)
        base_name = m.group(1)
    out_path = f"{base_name}_iq{iq}_wq{wq}_aq{aq}.h5"
    model_q.save(out_path, include_optimizer=False)
    print(f"Model successfully quantized and saved as {out_path}.")


def convert_model(model_path, input_scaling):
    """Wrapper to convert model"""
    base_name, ext = os.path.splitext(model_path)
    if ext.lower() == ".onnx":
        q_model = onnx.load_model(model_path)
    else:
        q_model = load_quantized_model(model_path)
    out_path = f"{base_name}.fbz"
    convert(q_model, input_scaling=input_scaling, file_path=out_path)
    print(f"Model successfully converted and saved as {out_path}.")


def reshape_model(model_path, input_height, input_width):
    """Wrapper to reshape model"""
    # Load the original model file
    base_model = load_quantized_model(model_path)

    # Extract base name
    base_name = os.path.splitext(model_path)[0]

    model = reshape(base_model, input_height, input_width)

    out_path = f"{base_name}_{input_height}_{input_width}.h5"

    model.save(out_path)
    print(f"Model input successfully reshaped and saved as {out_path}.")


def calibrate_adaround_model(base_model, base_name, samples, batch_size, epochs,
                             lr, include_act):
    """Wrapper to Adaround calibration model method"""

    out_path = f"{base_name}_adaround_calibrated.h5"

    # Apply Adaround
    optimizer = tf.keras.optimizers.Adam(learning_rate=lr)
    adaround_model = adaround(base_model,
                              samples,
                              optimizer,
                              epochs=epochs,
                              loss=tf.keras.losses.MeanSquaredError(),
                              batch_size=batch_size,
                              include_activation=include_act)

    adaround_model.save(out_path)
    print(f"Model successfully calibrated with Adaround method and saved \
            as {out_path}.")


def calibrate_bias_correction_model(base_model, base_name, samples, batch_size):
    """Wrapper to Bias Correction calibration model method"""

    out_path = f"{base_name}_bc_calibrated.h5"

    # Apply bias correction
    bc_model = bias_correction(base_model, samples, batch_size)

    bc_model.save(out_path)
    print(f"Model successfully calibrated with Bias Correction method and \
            saved as {out_path}.")


def main():
    """CNN2SNN command-line interface to quantize/convert/upgrade a model"""
    parser = argparse.ArgumentParser()
    parent_parser = argparse.ArgumentParser(add_help=False)
    parent_parser.add_argument("-m",
                               "--model",
                               type=str,
                               required=True,
                               help="The source model")
    sp = parser.add_subparsers(dest="action")
    q_parser = sp.add_parser("quantize",
                             parents=[parent_parser],
                             help="Quantize a Keras model")
    q_parser.add_argument("-wq",
                          "--weight_quantization",
                          type=int,
                          default=4,
                          help="The weight quantization")
    q_parser.add_argument("-aq",
                          "--activ_quantization",
                          type=int,
                          default=4,
                          help="The activations quantization")
    q_parser.add_argument("-iq",
                          "--input_weight_quantization",
                          type=int,
                          default=-1,
                          help="The first layer weight quantization (same as"
                          " weight_quantization if omitted)")
    q_parser.add_argument("--no-fold-BN",
                          dest="fold_BN",
                          action='store_false',
                          help="If specified, do not fold BatchNormalization "
                          "layers before quantization")
    c_parser = sp.add_parser(
        "convert",
        parents=[parent_parser],
        help="Convert a quantized Keras/ONNX model to an Akida model")
    c_parser.add_argument("-sc",
                          "--scale",
                          type=int,
                          default=None,
                          help="The scale factor applied on uint8 inputs.")
    c_parser.add_argument("-sh",
                          "--shift",
                          type=int,
                          default=None,
                          help="The shift applied on uint8 inputs.")
    r_parser = sp.add_parser(
        "reshape",
        parents=[parent_parser],
        help="Reshape a (quantized) Keras model Input layer to a given size.")
    r_parser.add_argument("-ih",
                          "--input_height",
                          type=int,
                          default=None,
                          required=True,
                          help="The new input height.")
    r_parser.add_argument("-iw",
                          "--input_width",
                          type=int,
                          default=None,
                          required=True,
                          help="The new input width.")
    calib_parser = sp.add_parser("calibrate",
                                 help="Calibrate a quantized Keras model")
    # Create parent subparser for arguments shared between calibrate methods
    calib_parent = argparse.ArgumentParser(add_help=False,
                                           parents=[parent_parser])
    calib_parent.add_argument(
        "-sa",
        "--samples",
        type=str,
        default=None,
        required=True,
        help="Set of samples to calibrate the model parameters (.npz file).")
    calib_parent.add_argument("-bs",
                              "--batch_size",
                              type=int,
                              default=None,
                              help="The batch size.")
    calib_method_parser = calib_parser.add_subparsers(
        dest="method",
        help="Weight calibration method. Bias Correction or AdaRound.")
    calib_method_parser.add_parser(
        "bc",
        help="Bias Correction weight calibration method.",
        parents=[calib_parent])
    adaround_parser = calib_method_parser.add_parser(
        "adaround",
        help="Adaround weight calibration method.",
        parents=[calib_parent])
    adaround_parser.add_argument("-e",
                                 "--epochs",
                                 type=int,
                                 default=10,
                                 help="The number of epochs.")
    adaround_parser.add_argument("-lr",
                                 "--learning_rate",
                                 type=float,
                                 default=1e-2,
                                 help="The learning rate.")
    adaround_parser.add_argument("-act",
                                 "--include_activation",
                                 action='store_true',
                                 help="Include activation or not.")

    args = parser.parse_args()
    if args.action == "quantize":
        quantize_model(args.model, args.weight_quantization,
                       args.activ_quantization, args.input_weight_quantization,
                       args.fold_BN)
    if args.action == "convert":
        if args.scale is None or args.shift is None:
            input_scaling = None
        else:
            input_scaling = (args.scale, args.shift)
        convert_model(args.model, input_scaling)
    if args.action == "reshape":
        reshape_model(args.model, args.input_height, args.input_width)
    if args.action == "calibrate":
        data = np.load(args.samples)
        lst = data.files
        samples_arr = data[lst[0]]
        for item in lst[1:]:
            samples_arr = np.concatenate((samples_arr, data[item]))

        base_model = load_quantized_model(args.model)
        base_model = activations_rescaling(base_model, samples_arr,
                                           args.batch_size)
        base_model = normalize_separable_model(base_model)

        # Extract base name
        base_name = os.path.splitext(args.model)[0]

        if args.method == "adaround":
            calibrate_adaround_model(base_model, base_name, samples_arr,
                                     args.batch_size, args.epochs,
                                     args.learning_rate,
                                     args.include_activation)
        else:
            calibrate_bias_correction_model(base_model, base_name, samples_arr,
                                            args.batch_size)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert QuantizedStatefulRecurrent to Akida.
"""
from akida import StatefulRecurrent
import quantizeml.layers as qlayers
import numpy as np

from ..akida_versions import AkidaVersion
from .weights import broadcast_and_set_variable
from .outputs import set_output_v2_variables, parse_output_bits, parse_post_op_buffer_bits
from .block_converter import BlockConverter, register_conversion_patterns
from .blocks import get_block_out_quantizer
from .activations import set_relu_variables
from .conv_common import get_layer_by_type

__all__ = ["StatefulRecurrentBlockConverter"]

_PATTERNS = [(qlayers.QuantizedDense, qlayers.QuantizedStatefulRecurrent,
              qlayers.QuantizedExtractToken, qlayers.QuantizedDense, qlayers.QuantizedReLU),
             (qlayers.QuantizedDense, qlayers.QuantizedStatefulRecurrent,
              qlayers.QuantizedExtractToken, qlayers.QuantizedDense)]


def _set_stateful_recurrent_variables(ak_layer, block):
    """Computes and sets the variables for an Akida StatefulRecurrent layer.

    This function converts the variables of a Keras layer and sets them into
    the corresponding variables of the equivalent Akida layer.

    Args:
        ak_layer (:obj:`akida.Layer`): the targeted akida layer.
        block (list(:obj:`tf.keras.Layer`)): the block of keras layers.
    """
    variables_ak = ak_layer.variables

    in_proj_dense = block[0]
    stateful_rec = block[1]
    out_proj_dense = block[3]

    assert isinstance(in_proj_dense.weight_quantizer, qlayers.WeightQuantizer)
    assert isinstance(out_proj_dense.weight_quantizer, qlayers.WeightQuantizer)
    assert isinstance(stateful_rec.a_quantizer, qlayers.WeightQuantizer)

    # Set StatefulRecurrent input_shift
    input_shift = getattr(in_proj_dense, 'input_shift', None)
    if input_shift is not None:
        broadcast_and_set_variable(variables_ak, "input_shift",
                                   input_shift.value.numpy().astype(np.uint8))
    # Set input_projection dense kernel
    in_proj_ak = in_proj_dense.weight_quantizer.qweights.value.fp.values.numpy()
    variables_ak["in_proj"] = in_proj_ak.astype(np.int8)

    # Set intermediate input_proj scale_out scale variable
    if hasattr(in_proj_dense.out_quantizer, 'qscales'):
        scales = in_proj_dense.out_quantizer.qscales.value.values
        output_scales = scales.numpy().astype(np.uint8)
        broadcast_and_set_variable(variables_ak, "in_proj_outputs_scales", output_scales)

    # Set intermediate input_proj scale_out shift variable.
    # Note that it's merged with the StatefulRecurrent input shifts variables.
    stateful_inputs_shift_ak = 0
    stateful_inputs_shift = getattr(stateful_rec, 'input_shift', None)
    if stateful_inputs_shift is not None:
        stateful_inputs_shift_ak = stateful_inputs_shift.value.numpy()
    stateful_inputs_shift_ak += stateful_rec.input_add_shift.value.numpy()
    stateful_inputs_shift_ak += in_proj_dense.out_quantizer.shift.value.numpy()
    broadcast_and_set_variable(variables_ak, "input_proj_to_stateful_shift",
                               stateful_inputs_shift_ak.astype(np.int8))

    # Set A weights variables
    A_real_ak = stateful_rec.a_quantizer(stateful_rec.A_real).values.numpy()
    variables_ak["A_real"] = A_real_ak.astype(np.int16)
    A_imag_ak = stateful_rec.a_quantizer(stateful_rec.A_imag).values.numpy()
    variables_ak["A_imag"] = A_imag_ak.astype(np.int16)

    # Set intermediate stateful_outputs_shift
    internal_state_shift = stateful_rec.out_quantizer.shift.value[0]
    broadcast_and_set_variable(variables_ak, "internal_state_outputs_shift",
                               internal_state_shift.numpy().astype(np.int8))

    # Set out_projection input_shift
    out_proj_inputs_shift = getattr(out_proj_dense, 'input_shift', None)
    if out_proj_inputs_shift is not None:
        broadcast_and_set_variable(variables_ak, "out_proj_inputs_shift",
                                   out_proj_inputs_shift.value.numpy().astype(np.uint8))

    # Set out_proj dense kernel variable
    out_proj_ak = out_proj_dense.weight_quantizer.qweights.value.fp.values.numpy()
    variables_ak["out_proj"] = out_proj_ak.astype(np.int8)

    if out_proj_dense.use_bias:
        bias_quantizer = out_proj_dense.bias_quantizer
        assert isinstance(bias_quantizer, qlayers.AlignedWeightQuantizer)
        # Set StatefuRecurrent layer bias variable and shift
        bias = bias_quantizer.qweights.value.values.numpy().astype(np.int32)
        bias_shift = bias_quantizer.shift.value.numpy().astype(np.uint8)
        variables_ak["bias"] = (bias >> bias_shift).astype(np.int8)
        broadcast_and_set_variable(variables_ak, "bias_shift", bias_shift)

    # Check if we have ReLU
    relu_layer = get_layer_by_type(block, qlayers.QuantizedReLU)
    if relu_layer:
        # Set ReLU variables
        set_relu_variables(ak_layer, relu_layer)

    out_quantizer = get_block_out_quantizer(block)
    if out_quantizer:
        set_output_v2_variables(ak_layer, out_quantizer)


def _create_stateful_recurrent(block):
    """Parses a quantizeml quantized StatefulRecurrent layers block and returns the
    params to create the corresponding Akida StatefulRecurrent layer.

    Args:
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        :obj:`akida.StatefulRecurrent`: The created akida layer.
    """
    stateful_rec = block[1]
    out_proj_dense = block[3]
    if out_proj_dense.buffer_bitwidth != block[0].buffer_bitwidth:
        raise ValueError("in_proj and out_proj should have the same buffer_bitwidth")
    relu_layer = get_layer_by_type(block, qlayers.QuantizedReLU)
    # In quantizeml one bit is reserved automatically for the sign, but in
    # akida this is rather checked during the clipping operations.
    block_params = {"dense_buffer_bits": out_proj_dense.buffer_bitwidth + 1,
                    "buffer_bits": stateful_rec.buffer_bitwidth + 1,
                    "num_coeffs": stateful_rec.num_coeffs,
                    "repeat": stateful_rec.repeat,
                    "subsample_ratio": stateful_rec.subsample_ratio,
                    "out_channels": out_proj_dense.units,
                    "activation": True if relu_layer else False}

    # parse the block output bits
    parse_output_bits(block, block_params)
    # parse the block post op buffer bits
    parse_post_op_buffer_bits(block, block_params)

    return StatefulRecurrent(**block_params,
                             name=out_proj_dense.name)


def convert_quantized_statefulrec(model_ak, block):
    """Converts QuantizedLayerStatefulRecurrent layers block and its variables and adds
    it to the Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the model will be added.
        block (list(:obj:`tf.keras.Layer`)): list of quantizeml quantized layers.

    Returns:
        bool: Returns True for a successful conversion.
    """
    # Create and add layer to the akida model
    layer_ak = _create_stateful_recurrent(block)
    model_ak.add(layer_ak)
    # Set the akida layer converted variables
    _set_stateful_recurrent_variables(layer_ak, block)
    return True


class StatefulRecurrentBlockConverter(BlockConverter):
    """Main class that should be used to check if the stateful recurrent layer block is compatible to an
    Akida v2 conversion and provides a method to convert it in an equivalent Akida v2 MadNorm layer.

    Args:
        block (list): list of quantizeml quantized layers.
    """

    def convert(self, model_ak):
        return convert_quantized_statefulrec(model_ak, self._block)


# Register the valid stateful recurrent block pattern for Akida v2
register_conversion_patterns(AkidaVersion.v2, _PATTERNS, StatefulRecurrentBlockConverter)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
__all__ = ["get_inbound_layers"]


def get_inbound_layers(ak_model, node, graph):
    """Returns a list of Akida layers in an Akida model that correspond to the inbound
    nodes of a NodeProto.

    Args:
        ak_model (Model): the model where to find the inbounds layers.
        node (NodeProto): the node to get the inbounds.
        graph (GraphProto): the graph containing the node.

    Returns:
        list: the inbounds akida layers.
    """
    # Get the nodes connected to the inputs, keeping the same order
    inbound_nodes = [""] * len(node.input)
    inodes_names = list(node.input)
    for t_node in graph.node:
        try:
            inbound_nodes[inodes_names.index(t_node.output[0])] = t_node
        except ValueError:
            pass
    inbound_nodes = list(filter(lambda x: x, inbound_nodes))

    # Check each inbound has a name
    if not all([node.name for node in inbound_nodes]):
        raise ValueError("All inbound nodes must have a name.")

    # Get the corresponding Akida layers
    return [ak_model.get_layer(node.name) for node in inbound_nodes]



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to update akida layer output variables from a keras OutputQuantizer.
"""
import numpy as np
import tensorflow as tf
from quantizeml.layers import OutputQuantizer
from quantizeml.tensors import pow2
from .weights import broadcast_and_set_variable
from .blocks import get_block_out_quantizer


def parse_output_bits(block, block_params):
    """ Helper to set the block output bits in its corresponding dict block_params.
    Note that this helper should be called only once the buffer_bits is set. What is normally done
    when the main block layer is parsed.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the layers block.
        block_params (dict): the current block parameters.
    """
    # Get the layer block output_quantizer
    out_quantizer = get_block_out_quantizer(block)
    if out_quantizer:
        block_params.update({"output_bits": out_quantizer.bitwidth})
    else:
        # Default to buffer bitwidth
        block_params.update({'output_bits': block_params["buffer_bits"]})


def parse_post_op_buffer_bits(block, block_params):
    """ Helper to set the block post op buffer bits in its corresponding dict block_params.

    Args:
        block (list(:obj:`tf.keras.Layer`)): the layers block.
        block_params (dict): the current block parameters.
    """
    # Get the layer block output_quantizer
    out_quantizer = get_block_out_quantizer(block)
    if out_quantizer:
        block_params.update({"post_op_buffer_bits": out_quantizer.buffer_bitwidth})


def set_output_v1_variables(layer_ak, out_quantizer):
    """Computes and sets the output variables in an akida v1 layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida v1 layer.
        out_quantizer (:obj:`quantizeml.OutputQuantizer`): the source output quantizer.
    """
    assert isinstance(out_quantizer, OutputQuantizer)

    # Extract the OutputQuantizer variables
    scales = out_quantizer.qscales.value.values
    shift = out_quantizer.shift.value

    # Quantizeml evaluates the outputs as: y = x * scales * 2^shift
    # Calculate the float activation step, as the reciprocal of (scales * 2^shift)
    scales_rec = (pow2(-shift) / scales).numpy().astype(np.float32)
    # In akida 1.0, the outputs are evaluated as: y = x / act_step
    layer_ak.variables["act_step"] = scales_rec
    if layer_ak.parameters.activation:
        # For activations, x is decreased by half the activations step before division
        # to increase the precision. This is obtained by increasing the threshold.
        layer_ak.variables["threshold"] += np.round(0.5 * scales_rec).astype(np.int32)
        # Adjust activation step to match activation formula
        act_bits = layer_ak.parameters.act_bits
        layer_ak.variables["act_step"] *= 2 ** (act_bits - 4)


def set_output_v2_variables(layer_ak, out_quantizer):
    """Computes and sets the output variables in an akida v2 layer.

    Args:
        layer_ak (:obj:`akida.Layer`): the targeted akida v2 layer.
        out_quantizer (:obj:`quantizeml.OutputQuantizer`): the source output quantizer.
    """
    assert isinstance(out_quantizer, OutputQuantizer)

    # Extract the OutputQuantizer variables.
    # Note that qscales exist only if the out_quantizer was used on QFloat (not FixedPoint as in
    # the QuantizedAdd layer)
    if hasattr(out_quantizer, 'qscales'):
        scales = out_quantizer.qscales.value.values
        output_scales = scales.numpy().astype(np.uint8)
        broadcast_and_set_variable(layer_ak.variables, "output_scales",
                                   output_scales)

    # clip shift value to ensure that it is within a valid range
    shift = tf.maximum(out_quantizer.shift.value, -layer_ak.parameters.buffer_bits + 1)
    broadcast_and_set_variable(layer_ak.variables, "output_shift",
                               shift.numpy().astype(np.int8))



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Equalization tools for Keras/CNN2SNN Sequential models.

   These transformations take models with heterogeneities in their weights or
   biases that are identified as harmful for the model quantization of for their
   conversion to akida. They produce models with reduced heterogeneities, yet
   returning nearly identical results.
"""

import numpy as np
import keras

from .clone import clone_model_with_weights


def _get_filter_max_values(layer):
    if isinstance(layer, keras.layers.Conv2D):
        # Layer "weights" are filters, biases
        filters = layer.get_weights()[0]
        # Filters are HWCN, i.e. the filter index is the last dimension
        return np.amax(np.abs(filters), axis=(0, 1, 2))
    if isinstance(layer, keras.layers.SeparableConv2D):
        # Layer "weights" are depthwise filters, pointwise filters, biases
        # We are only interested in rescaling pointwise filters
        filters = layer.get_weights()[1]
        # Pointwise filters are HWCN, i.e. the filter index is the last
        # dimension
        return np.amax(np.abs(filters), axis=(0, 1, 2))
    if isinstance(layer, keras.layers.Dense):
        # Layer "weights" are filters, biases
        filters = layer.get_weights()[0]
        # Filters are CN, i.e. the filter index is the last dimension
        return np.amax(np.abs(filters), axis=0)
    return None


def _get_channel_max_values(layer):
    if isinstance(layer, keras.layers.Conv2D):
        # Layer "weights" are filters, biases
        filters = layer.get_weights()[0]
        # Filters are HWCN, i.e. the channel index is the third dimension
        return np.amax(np.abs(filters), axis=(0, 1, 3))
    if isinstance(layer, keras.layers.SeparableConv2D):
        # Layer "weights" are depthwise filters, pointwise filters, biases
        dw_filters = layer.get_weights()[0]
        # Depthwise filters are HWCN, i.e. channel index is the third dimension
        return np.amax(np.abs(dw_filters), axis=(0, 1, 3))
    if isinstance(layer, keras.layers.Dense):
        # Layer "weights" are filters, biases
        filters = layer.get_weights()[0]
        # Filters are CN, i.e. the channel index is the first dimension
        return np.amax(np.abs(filters), axis=1)
    return None


def _rescale_fused_layer(layer, scales):
    weights = layer.get_weights()
    if isinstance(layer, keras.layers.SeparableConv2D):
        # Layer "weights" are depthwise filters, pointwise filters, biases
        dw_filters = weights[0]
        pw_filters = weights[1]
        # Divide dw_filters and multiply pw_filters by the same scale
        for i in range(scales.shape[0]):
            dw_filters[:, :, i, :] = dw_filters[:, :, i, :] / scales[i]
            pw_filters[:, :, i, :] = pw_filters[:, :, i, :] * scales[i]
        weights[0] = dw_filters
        weights[1] = pw_filters
        # Update layer weights
        layer.set_weights(weights)


def _get_homogeneity_rate(layer):
    filter_max = _get_filter_max_values(layer)
    if filter_max is None:
        return None
    global_max = np.max(filter_max)
    # Evaluate the relative score of each filter
    filter_rate = filter_max / global_max
    return np.mean(filter_rate)


def weights_homogeneity(model):
    """Give an estimation of the homogeneity of layer weights

    For each Conv or Dense layer in the model, this compares the ranges of
    the weights for each filter with the range of the tensor.
    The score for each filter is expressed as an homogeneity rate (1 is the
    maximum), and the layer homogeneity rate is the mean of all filter rates.

    Args:
        model (:obj:`tf.keras.Model`): a Keras model.
    Returns:
        dict: rates indexed by layer names.
    """
    scores = {}
    for layer in model.layers:
        score = _get_homogeneity_rate(layer)
        if score is not None:
            scores[layer.name] = score
    return scores


def normalize_separable_layer(layer):
    """This normalizes the depthwise weights of a SeparableConv2D.

       In order to limit the quantization error when using a per-tensor
       quantization of depthwise weights, this rescales all depthwise weights
       to fit within the [-1, 1] range.
       To preserve the output of the layer, each depthwise kernel is rescaled
       independently to the [-1, 1] interval by dividing all weights by the
       absolute maximum value, and inversely, all pointwise filters 'looking'
       at these kernels are multiplied by the same value.

    Args:
        layer (:obj:`tf.keras.layers.SeparableConv2D`): a Keras SeparableConv2D
            layer.
    """
    if not isinstance(layer, keras.layers.SeparableConv2D):
        raise ValueError("The layer is not a SeparableConv2D")
    # Get maximum ranges per channel
    dw_max = _get_channel_max_values(layer)
    # Rescale depthwise to [-1, 1] and adjust pointwise accordingly
    _rescale_fused_layer(layer, dw_max)


def normalize_separable_model(model):
    """This normalizes the depthwise weights of all SeparableConv2D in a Model.

    Args:
        model (:obj:`tf.keras.Model`): a Keras model.

    Returns:
        :obj:`tf.keras.Model`: a new Keras model with normalized depthwise
        weights in SeparableConv2D layers.
    """
    # Clone model
    new_model = clone_model_with_weights(model)

    # Normalize SeparableConv2D depthwise weights
    for layer in new_model.layers:
        if isinstance(layer, keras.layers.SeparableConv2D):
            normalize_separable_layer(layer)

    return new_model



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""Functions to convert input layers to Akida layers.
"""
from keras.layers import InputLayer
from keras import Sequential
from akida import Model, InputData
from quantizeml.layers import StatefulRecurrent
from .conv_common import get_layer_by_type
from ..akida_versions import AkidaVersion, get_akida_version


def _create_input_data(layer, input_16bits=False):
    """Parses a keras InputLayer layer and returns the corresponding Akida
    InputData layer.

    Args:
        layer (:obj:`tf.keras.Layer`): the InputLayer to convert.
        input_16bits (bool, optional): boolean to specify that the inputs type is 16bit or not.
            Defaults to False.

    Returns:
        akida.InputData: The parsed layer.
    """
    if not isinstance(layer, InputLayer):
        raise TypeError(f"Layer {layer.name} was expected to be "
                        "InputLayer")
    # input_shape param is stored as a tuples in an array
    input_shape = layer.input_shape[0]

    if input_shape[0] is not None:
        raise ValueError("First dimension for input layer expected to be None")
    # Ignore first dimension, because it is None
    input_shape = input_shape[1:]
    # input shape must not exceeds 3 dim
    if len(input_shape) > 3:
        raise ValueError("input shape must not exceed 3 dimensions."
                         f"Receives {input_shape}")

    # Create a list with 1 for each dimension that we miss before having 3
    # dims, that is what akida expects
    missing_dimensions = [1] * (3 - len(input_shape))
    input_shape = [*missing_dimensions, *input_shape]
    # the akida version determines the input_bits and input sign
    if get_akida_version() == AkidaVersion.v1:
        input_bits = 4
        input_signed = False
    else:
        input_bits = 16 if input_16bits else 8
        input_signed = True
    # With the convert api of cnn2snn the model should handle only unsigned inputs.
    return InputData(input_shape=input_shape,
                     input_bits=input_bits,
                     input_signed=input_signed,
                     name=layer.name)


def convert_input(model_ak, model_k):
    """Converts a keras Input layer into ``akida.InputData`` and add it to the
    Akida's model.

    Args:
        model_ak (:obj:`akida.Model`): the Akida model where the Akida InputData will be inserted.
        model_k (:obj:`keras.Layer`): the keras quantized model.
    """
    if not isinstance(model_ak, Model):
        raise TypeError(f"Expecting an akida model, received {type(model_ak)}")
    if len(model_ak.layers) != 0:
        raise TypeError("Akida model should be empty")

    # If the quantized keras model is Sequential, it does not have an InputLayer. Hence the
    # input_shape should be extracted from its first neural layer.
    if isinstance(model_k, Sequential):
        input_shape = getattr(model_k.layers[0], "input_shape", None)
        assert input_shape is not None, "input_shape not found"
        layer_k = InputLayer(input_shape=input_shape[1:])
    else:
        layer_k = model_k.layers[0]
        if not isinstance(layer_k, InputLayer):
            raise TypeError(f"Layer {layer_k.name} was expected to be InputLayer.")

    stateful_layers = get_layer_by_type(model_k.layers, StatefulRecurrent)
    input_data = _create_input_data(layer_k, stateful_layers is not None)
    model_ak.add(input_data)
